#
# (C) Copyright 2017- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import copy
import datetime
import logging
import os
from pathlib import Path

from . import utils
import numpy as np
import pandas as pd
import yaml

# logging.basicConfig(level=logging.DEBUG)
# logging.basicConfig(level=logging.INFO)
LOG = logging.getLogger(__name__)

NEWER = True


class GribIndexer:
    VECTOR_PARAMS = {
        "wind10m": ["10u", "10v"],
        "wind100m": ["100u", "100v"],
        "wind200m": ["200u", "200v"],
        "wind": ["u", "v"],
        "wind3d": ["u", "v", "w"],
    }

    # tuple-> 0: ecCodes type, 1: pandas type, 2: Python type 3: use in duplicate check
    DEFAULT_KEYS = {
        "shortName": ("s", str, str, False),
        "paramId": ("l", "Int32", int, False),
        "date": ("l", "Int64", int, True),
        "time": ("l", "Int64", int, True),
        "step": ("l", "Int32", int, True),
        "level": ("l", "Int32", int, True),
        "typeOfLevel": ("s", str, str, False),
        "number": ("s", str, str, True),
        "experimentVersionNumber": ("s", str, str, False),
        "marsClass": ("s", str, str, False),
        "marsStream": ("s", str, str, False),
        "marsType": ("s", str, str, False),
    }

    DEFAULT_ECC_KEYS = [f"{k}:{v[0]}" for k, v in DEFAULT_KEYS.items()]
    BLOCK_KEYS = ["shortName", "typeOfLevel"]

    DEFAULT_SORT_KEYS = [
        "date",
        "time",
        "step",
        "number",
        "level",
        "paramId",
    ]
    DATE_KEYS = {
        k: ("l", "Int64", int)
        for k in ["date", "dataDate", "validityDate", "mars.date", "marsDate"]
    }
    TIME_KEYS = {
        k: ("l", "Int64", int)
        for k in ["time", "dataTime", "validityTime", "mars.time", "marsTime"]
    }

    DATETIME_KEYS = {
        "_dateTime": ("date", "time"),
        "_dataDateTime": ("dataDate", "dataTime"),
        "_validityDateTime": ("validityDate", "validityTime"),
    }

    KEYS_TO_REPLACE = {
        ("type", "mars.type"): "marsType",
        ("stream", "mars.stream"): "marsStream",
        ("class", "mars.class", "class_"): "marsClass",
        ("perturbationNumber"): "number",
        ("mars.date", "marsDate"): "date",
        ("mars.time", "marsTime"): "time",
    }

    PREDEF_KEYS = copy.deepcopy(DEFAULT_KEYS)
    PREDEF_KEYS.update(DATE_KEYS)
    PREDEF_KEYS.update(TIME_KEYS)

    PREDEF_PD_TYPES = {k: v[1] for k, v in PREDEF_KEYS.items()}
    PREDEF_PT_TYPES = {k: v[2] for k, v in PREDEF_KEYS.items()}

    def __init__(self, db):
        self.db = db
        assert self.db is not None
        self.ref_column_count = None
        self.keys = []
        self.keys_ecc = []
        for k, v in GribIndexer.DEFAULT_KEYS.items():
            name = k
            self.keys.append(name)
            if v[0]:
                name = f"{k}:{v[0]}"
            self.keys_ecc.append(name)

        self.keys_duplicate_check = [
            k for k, v in GribIndexer.DEFAULT_KEYS.items() if v[3] == True
        ]

        self.shortname_index = self.keys.index("shortName")
        self.levtype_index = self.keys.index("typeOfLevel")
        self.type_index = self.keys.index("marsType")
        self.number_index = self.keys.index("number")
        self.param_id_index = self.keys.index("paramId")

        self.wind_check_index = []
        for v in [
            "date",
            "time",
            "step",
            "level",
            "typeOfLevel",
            "level",
            "number",
            "experimentVersionNumber",
            "marsClass",
            "marsStream",
            "marsType",
        ]:
            self.wind_check_index.append(self.keys.index(v) + 1)

        # self.block_key_index = [self.keys.index(v) for v in GribIndexer.BLOCK_KEYS]

        self.pd_types = {k: v[1] for k, v in GribIndexer.DEFAULT_KEYS.items()}
        self.pt_types = {k: v[2] for k, v in GribIndexer.DEFAULT_KEYS.items()}

    def update_keys(self, keys):
        ret = False
        for k in keys:
            name = k
            # we do not add datetime keys (they are pseudo keys, and their value
            # is always generated on the fly)
            if name not in self.keys and name not in GribIndexer.DATETIME_KEYS:
                self.keys.append(name)
                p = GribIndexer.PREDEF_KEYS.get(name, ("", str, str))
                ecc_name = name if p[0] == "" else name + ":" + p[0]
                self.keys_ecc.append(ecc_name)
                self.pd_types[name] = p[1]
                self.pt_types[name] = p[2]
                ret = True
        return ret

    def _check_duplicates(self, name, df):
        dup = df.duplicated(subset=self.keys_duplicate_check)
        first_dup = True
        cnt = 0
        for i, v in dup.items():
            if v:
                if first_dup:
                    LOG.error(
                        f"{name}: has duplicates for key group: {self.keys_duplicate_check}!"
                    )
                    first_dup = False
                    LOG.error(f" first duplicate: {df.iloc[i]}")
                cnt += 1

        if cnt > 1:
            LOG.error(f"  + {cnt-1} more duplicate(s)!")

    def _build_vector_index(self, df, v_name, v_comp):
        # LOG.debug(f"v_name={v_name} v_comp={v_comp}")
        comp_num = len(v_comp)

        # filter components belonging together
        comp_df = []
        for i, comp_name in enumerate(v_comp):
            query = f"shortName == '{comp_name}'"
            r = df.query(query, engine="python")
            # if we do not use copy, the assignment below as:
            # comp_df[0].loc[...
            # generates the SettingWithCopyWarning warning!!!
            if i == 0:
                r = df.query(query, engine="python").copy()
            else:
                r = df.query(query, engine="python")
            if r.empty:
                return []
            else:
                comp_df.append(r)

        assert comp_num == len(comp_df)

        # pair up components within a 2D vector field. This
        # version proved to be the fastest!
        # LOG.debug(" pair up collected components:")

        # print(f"v_name={v_name} {len(comp_df[1].index)}")
        # print("view=", comp_df[0]._is_view)
        r = []
        used1 = np.full(len(comp_df[1].index), False, dtype="?")
        comp_df[0].loc[:, "shortName"] = v_name
        # 2D
        if comp_num == 2:
            for row0 in comp_df[0].itertuples(name=None):
                i = 0
                for row1 in comp_df[1].itertuples(name=None):
                    if not used1[i]:
                        b = True
                        for x in self.wind_check_index:
                            if row0[x] != row1[x]:
                                b = False
                                break
                        if b:
                            d = list(row0[1:])
                            d.extend(row1[-self.ref_column_count :])
                            r.append(d)
                            used1[i] = True
                            break
                    i += 1
        # 3D
        elif comp_num == 3:
            used2 = np.full(len(comp_df[2].index), False, dtype="?")
            for row0 in comp_df[0].itertuples(name=None):
                i = 0
                for row1 in comp_df[1].itertuples(name=None):
                    if not used1[i]:
                        b = True
                        for x in self.wind_check_index:
                            if row0[x] != row1[x]:
                                b = False
                                break
                        if b:
                            j = 0
                            for row2 in comp_df[2].itertuples(name=None):
                                if not used2[j]:
                                    b = True
                                    for x in self.wind_check_index:
                                        if row0[x] != row2[x]:
                                            b = False
                                            break
                                    if b:
                                        d = list(row0[1:])
                                        d.extend(row1[-self.ref_column_count :])
                                        d.extend(row2[-self.ref_column_count :])
                                        r.append(d)
                                        used1[i] = True
                                        used2[j] = True
                                        j = -1
                                        break
                                j += 1
                            if j == -1:
                                break
                    i += 1
        return r

    def _make_dataframe(self, data, sort=False, columns=None):
        if columns is not None:
            df = pd.DataFrame(data, columns=columns)
        else:
            df = pd.DataFrame(data)

        for c in df.columns:
            if self.pd_types.get(c, "") in ["Int32", "Int64"]:
                df.fillna(value={c: np.nan}, inplace=True)
            df = df.astype(self.pd_types)
        if sort:
            df = GribIndexer._sort_dataframe(df)

        return df

    @staticmethod
    def _sort_dataframe(df, columns=None, ascending=True):
        if columns is None:
            columns = list(df.columns)
        elif not isinstance(columns, list):
            columns = [columns]

        # mergesoft is a stable sorting algorithm
        df = df.sort_values(by=columns, ascending=ascending, kind="mergesort")
        df = df.reset_index(drop=True)
        return df

    def _write_dataframe(self, df, name, out_dir):
        f_name = os.path.join(out_dir, f"{name}.csv.gz")
        df.to_csv(path_or_buf=f_name, header=True, index=False, compression="gzip")

    @staticmethod
    def read_dataframe(key, dir_name):
        # assert len(key) == len(GribIndexer.BLOCK_KEYS)
        name = key
        f_name = os.path.join(dir_name, f"{name}.csv.gz")
        # LOG.debug("f_name={}".format(f_name))
        return pd.read_csv(f_name, index_col=None, dtype=GribIndexer.PREDEF_PD_TYPES)

    @staticmethod
    def get_storage_key_list(dir_name):
        r = []
        # LOG.debug(f"dir_name={dir_name}")
        suffix = ".csv.gz"
        for f in utils.get_file_list(os.path.join(dir_name, f"*{suffix}")):
            name = os.path.basename(f)
            # LOG.debug(f"name={name}")
            r.append(name[: -len(suffix)])
        return r

    @staticmethod
    def is_key_wind(key):
        return key in GribIndexer.VECTOR_PARAMS

    @staticmethod
    def _convert_query_value(v, col_type):
        # print(f"v={v} {type(v)} {col_type}")
        return v if col_type != "object" else str(v)

    @staticmethod
    def _check_datetime_in_filter_input(keys):
        for k, v in GribIndexer.DATETIME_KEYS.items():
            name = k[1:]
            name_date = v[0]
            name_time = v[1]
            if keys.get(name, []) and (
                keys.get(name_date, []) or keys.get(name_time, [])
            ):
                raise Exception(
                    f"Cannot specify {name} together with {name_date} and {name_time}!"
                )

    @staticmethod
    def _convert_filter_value(name, val):
        """
        Analyse the filter key-value pairs and perform the necessary conversions
        """
        valid_name = name.split(":")[0] if ":" in name else name

        # datetime keys are pseudo keys, they start with _. Their value is converted to
        # datetime. The key itself is not added to the scan!
        if ("_" + valid_name) in GribIndexer.DATETIME_KEYS:
            valid_name = "_" + valid_name
            name_date = GribIndexer.DATETIME_KEYS[valid_name][0]
            name_time = GribIndexer.DATETIME_KEYS[valid_name][1]
            for i, t in enumerate(val):
                val[i] = GribIndexer._to_datetime(name, t)
                # print(f"t={t} -> {val[i]}")
            # We add the date and time components with an empty value. So they will be
            # added to the scan, but they will be ignored by the query. Conversely,
            # the datetime key itself will be ignored in the scan, but will be used
            # in the query.
            return [("_" + name, val), (name_date, []), (name_time, [])]
        # we convert dates to int
        elif valid_name in GribIndexer.DATE_KEYS:
            for i, t in enumerate(val):
                d = GribIndexer._to_date(name, t)
                # for daily climatologies dates where the year is missing the
                # the a tuple is returned
                if not isinstance(d, tuple):
                    val[i] = int(d.strftime("%Y%m%d"))
                else:
                    val[i] = d[0] * 100 + d[1]
        # we convert times to int
        elif valid_name in GribIndexer.TIME_KEYS:
            for i, t in enumerate(val):
                val[i] = int(GribIndexer._to_time(name, t).strftime("%H%M"))
                # print(f"t={t} -> {val[i]}")
        else:
            pt_type = GribIndexer.PREDEF_PT_TYPES.get(name, None)
            # print(f"name={name} {pt_type}")
            if pt_type is not None:
                for i, t in enumerate(val):
                    val[i] = pt_type(t)
                    # print(f" t={t} -> {val[i]}")

        # remap some names to the ones already in the default set of indexer keys
        for k, v in GribIndexer.KEYS_TO_REPLACE.items():
            if name in k:
                name = v

        return [(name, val)]

    @staticmethod
    def _to_datetime(param, val):
        try:
            if isinstance(val, datetime.datetime):
                return val
            elif isinstance(val, str):
                return utils.date_from_str(val)
            elif isinstance(val, (int, float)):
                return utils.date_from_str(str(val))
            else:
                raise
        except:
            raise Exception(f"Invalid datetime value={val} specified for key={param}")

    @staticmethod
    def _to_date(param, val):
        try:
            if isinstance(val, datetime.datetime):
                return val.date()
            elif isinstance(val, datetime.date):
                return val
            elif isinstance(val, str):
                d = utils.date_from_str(val)
                return d.date() if not isinstance(d, tuple) else d
            elif isinstance(val, (int, float)):
                d = utils.date_from_str(str(val))
                return d.date() if not isinstance(d, tuple) else d
            else:
                raise
        except:
            raise Exception(f"Invalid date value={val} specified for key={param}")

    @staticmethod
    def _to_time(param, val):
        try:
            if isinstance(val, (datetime.datetime)):
                return val.time()
            elif isinstance(val, datetime.time):
                return val
            elif isinstance(val, str):
                return utils.time_from_str(val)
            elif isinstance(val, int):
                return utils.time_from_str(str(val))
            else:
                raise
        except:
            raise Exception(f"Invalid time value={val} specified for key={param}")


class FieldsetIndexer(GribIndexer):
    def __init__(self, *args):
        super().__init__(*args)
        self.ref_column_count = 1

    def scan(self, vector=False):
        data = self._scan(self.db.fs, mapped_params=self.db.mapped_params)
        if data:
            df = self._make_dataframe(data, sort=False)
            self.db.blocks["scalar"] = df
            if vector:
                self._scan_vector()

    def _scan(self, fs, mapped_params={}):
        LOG.info(f" scan fields ...")
        data = {}
        # print(f"fs_len={len(fs)}")
        # print(f"keys_ecc={self.keys_ecc}")
        if utils.is_fieldset_type(fs) and len(fs) > 0:
            md_vals = fs.grib_get(self.keys_ecc, "key")
            if mapped_params:
                for i in range(len(fs)):
                    v = md_vals[self.param_id_index][i]
                    if v in mapped_params:
                        short_name = mapped_params[v]
                        md_vals[self.shortname_index][i] = short_name

            assert len(self.keys) == len(self.keys_ecc)
            data = {k: md_vals[i] for i, k in enumerate(self.keys)}
            data["_msgIndex1"] = list(range(len(fs)))
            LOG.info(f" {len(fs)} GRIB messages processed")
        return data

    def _scan_vector(self):
        df = self.db.blocks["scalar"]
        if df is not None and not df.empty:
            for v_name, v_comp in GribIndexer.VECTOR_PARAMS.items():
                r = self._build_vector_index(df, v_name, v_comp)
                comp_num = len(v_comp)
                if r:
                    cols = [*self.keys]
                    for i in range(comp_num):
                        cols.extend([f"_msgIndex{i+1}"])
                    w_df = self._make_dataframe(r, sort=False, columns=cols)
                    self.db.blocks[v_name] = w_df
                    # self._write_dataframe(w_df, v_name, out_dir)
                else:
                    LOG.debug(" No paired fields found!")
                    continue


class ExperimentIndexer(GribIndexer):
    def __init__(self, *args):
        super().__init__(*args)
        self.ref_column_count = 2

    def scan(self):
        out_dir = self.db.db_dir
        Path(out_dir).mkdir(exist_ok=True, parents=True)
        LOG.info(f"scan {self.db} out_dir={out_dir} ...")

        data = {k: [] for k in [*self.keys, "_msgIndex1", "_fileIndex1"]}
        input_files = []

        # print(f"out_dir={out_dir}")
        # merge existing experiment objects
        if self.db.merge_conf:
            ds = []
            # simple merge
            if isinstance(self.db.merge_conf, list):
                for c_name in self.db.merge_conf:
                    ds.append(
                        {"data": db.dataset.find(c_name), "name": c_name, "ens": {}}
                    )
            # explicit ENS merge
            else:
                assert "pf" in self.db.merge_conf
                # control forecast
                c_name = self.db.merge_conf.get("cf", "")
                if c_name != "":
                    ds.append(
                        {
                            "data": self.db.dataset.find(c_name, comp="field"),
                            "name": c_name,
                            "ens": {"type": "cf", "number": 0},
                        }
                    )
                for i, c_name in enumerate(self.db.merge_conf.get("pf", [])):
                    ds.append(
                        {
                            "data": self.db.dataset.find(c_name, comp="field"),
                            "name": c_name,
                            "ens": {"type": "pf", "number": i + 1},
                        }
                    )

            for c in ds:
                if c["data"] is None:
                    c_name = d["name"]
                    raise Exception(
                        f"Cannot merge experiments as {self.db}! Experiment {c_name} is not found!"
                    )
                else:
                    input_files = self._scan_one(
                        input_dir=c["data"].path,
                        file_name_pattern=c["data"].file_name_pattern,
                        input_files=input_files,
                        mapped_params=self.db.mapped_params,
                        ens=c["ens"],
                        data=data,
                        rootdir_placeholder_value=c["data"].rootdir_placeholder_value,
                        rootdir_placeholder_token=self.db.ROOTDIR_PLACEHOLDER_TOKEN,
                    )
        # index a single experiment
        else:
            input_files = self._scan_one(
                input_dir=self.db.path,
                file_name_pattern=self.db.file_name_pattern,
                input_files=[],
                mapped_params=self.db.mapped_params,
                ens={},
                data=data,
                rootdir_placeholder_value=self.db.rootdir_placeholder_value,
                rootdir_placeholder_token=self.db.ROOTDIR_PLACEHOLDER_TOKEN,
            )

        # print(f"input_files={input_files}")
        if len(input_files) > 0 and len(data["shortName"]) > 0:
            # write config file for input file list
            LOG.info(f"generate datafiles.yaml ...")
            f_name = os.path.join(out_dir, "datafiles.yaml")
            r = yaml.dump(input_files, default_flow_style=False)
            with open(f_name, "w") as f:
                f.write(r)
            self.db.input_files = input_files

            # scalar
            LOG.info(f"generate scalar fields index ...")
            df = self._make_dataframe(data, sort=True)
            self.db.blocks["scalar"] = df
            self._write_dataframe(df, "scalar", out_dir)

            # vector (2D)
            LOG.info(f"generate vector fields index ...")
            for v_name, v_comp in GribIndexer.VECTOR_PARAMS.items():
                r = self._build_vector_index(df, v_name, v_comp)
                comp_num = len(v_comp)
                if r:
                    cols = [*self.keys]
                    for i in range(comp_num):
                        cols.extend([f"_msgIndex{i+1}", f"_fileIndex{i+1}"])
                    w_df = self._make_dataframe(r, sort=True, columns=cols)
                    # print(f"wind_len={len(w_df.index)}")
                    self.db.blocks[v_name] = w_df
                    self._write_dataframe(w_df, v_name, out_dir)
                else:
                    LOG.debug(" No paired fields found!")
                    continue

    def _scan_one(
        self,
        input_dir="",
        file_name_pattern="",
        input_files=[],
        mapped_params={},
        ens={},
        data={},
        rootdir_placeholder_value="",
        rootdir_placeholder_token=None,
    ):
        LOG.info("scan fields ...")
        LOG.info(f" input_dir={input_dir} file_name_pattern={file_name_pattern}")
        # print(f" input_dir={input_dir} file_name_pattern={file_name_pattern}")

        # for f_path in glob.glob(f_pattern):
        cnt = 0
        input_files_tmp = []
        for f_path in utils.get_file_list(
            input_dir, file_name_pattern=file_name_pattern
        ):
            # LOG.debug(f"  f_path={f_path}")
            fs = self.db.fieldset_class(path=f_path)
            if utils.is_fieldset_type(fs) and len(fs) > 0:
                cnt += 1
                input_files_tmp.append(f_path)
                file_index = len(input_files) + len(input_files_tmp) - 1
                md_vals = fs.grib_get(self.keys_ecc, "key")

                if mapped_params:
                    for i in range(len(fs)):
                        v = md_vals[self.param_id_index][i]
                        if v in mapped_params:
                            short_name = mapped_params[v]
                            md_vals[self.shortname_index][i] = short_name
                if ens:
                    for i in range(len(fs)):
                        md_vals[self.type_index][i] = ens["type"]
                        md_vals[self.number_index][i] = ens["number"]

                assert len(self.keys) == len(self.keys_ecc)
                for i, c in enumerate(self.keys):
                    data[c].extend(md_vals[i])
                data["_msgIndex1"].extend(list(range(len(fs))))
                data["_fileIndex1"].extend([file_index] * len(fs))

                # print({k: len(v) for k, v in data.items()})

        if rootdir_placeholder_value:
            input_files_tmp = [
                x.replace(rootdir_placeholder_value, rootdir_placeholder_token)
                for x in input_files_tmp
            ]

        input_files.extend(input_files_tmp)

        LOG.info(f" {cnt} GRIB files processed")
        return input_files

    def allowed_keys(self):
        r = list(self.keys)
        r.extend(GribIndexer.DATE_KEYS)
        r.extend(GribIndexer.TIME_KEYS)
        r.extend(list(GribIndexer.DATETIME_KEYS.keys()))
        return set(r)



#
# (C) Copyright 2017- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from datetime import datetime

import metview as mv
import pandas as pd

from metview.style import Style


class Track:
    def __init__(
        self,
        path,
        skiprows=None,
        sep=None,
        date_index=None,
        time_index=None,
        lat_index=None,
        lon_index=None,
    ):
        self.path = path
        self.skiprows = 0 if skiprows is None else skiprows
        self.sep = sep
        if self.sep == " ":
            self.sep = r"\s+"
        self.date_index = 0 if date_index is None else date_index
        self.time_index = 1 if time_index is None else time_index
        self.lat_index = 3 if lat_index is None else lat_index
        self.lon_index = 2 if lon_index is None else lon_index

    def style(self):
        return mv.style.get_db().get_style("track").clone()

    def build(self, style=None):
        style = [] if style is None else style
        df = pd.read_csv(
            filepath_or_buffer=self.path,
            sep=self.sep,
            skiprows=self.skiprows,
            header=None,
            engine="python",
        )

        # print(df)

        v_date = df.iloc[:, self.date_index]
        v_time = df.iloc[:, self.time_index]
        val = [" {}/{:02d}".format(str(d)[-2:], t) for d, t in zip(v_date, v_time)]
        lon = df.iloc[:, self.lon_index].values
        lat = df.iloc[:, self.lat_index].values
        idx_val = list(range(len(val)))

        # print(f"lon={lon}")
        # print(f"lat={lat}")
        # print(f"val={val}")
        # for x in style:
        #     print(f"style={x}")

        if len(style) == 0:
            s = mv.style.get_db().get_style("track").clone()
        if len(style) == 1 and isinstance(style[0], Style):
            s = style[0].clone()
        else:
            assert all(not isinstance(x, Style) for x in style)
            s = mv.style.get_db().get_style("track").clone()

        for vd in s.visdefs:
            if vd.verb == "msymb":
                vd.change_symbol_text_list(val, idx_val)

        r = s.to_request()

        vis = mv.input_visualiser(
            input_plot_type="geo_points",
            input_longitude_values=lon,
            input_latitude_values=lat,
            input_values=idx_val,
        )

        return [vis, *r]



#
# (C) Copyright 2017- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import datetime
import getpass
import glob
import logging
import math
from pathlib import Path
import shutil
import os
import re
import tempfile

import numpy as np

LOG = logging.getLogger(__name__)


CACHE_DIR = os.path.join(tempfile.gettempdir(), f"mpy-{getpass.getuser()}")


def deacc(fs, key=None, skip_first=False, mark_derived=False):
    r = None
    if key is None or key == "":
        if len(fs) > 1:
            v = fs[1:] - fs[:-1]
        if not skip_first:
            r = fs[0] * 0
            r = r.merge(v)
        else:
            r = v
    else:
        if not isinstance(key, str):
            raise TypeError(f"deacc(): key must be a str (got {type(key)})!")
        fs._get_db().load([key])
        key_vals = fs._unique_metadata(key)
        if key_vals:
            v = fs.select({key: key_vals[0]})
            gr_num = len(v)
            r = None
            if not skip_first:
                r = v * 0
            for i in range(1, len(key_vals)):
                v_next = fs.select({key: key_vals[i]})
                if len(v_next) != gr_num:
                    raise ValueError(
                        f"deacc(): unexpected number of fields (={len(v_next)}) found for {key}={key_vals[i]}! For each {key} value the number of fields must be the same as for {key}={key_vals[0]} (={gr_num})!"
                    )
                if r is None:
                    r = v_next - v
                else:
                    # print(f"i={i}")
                    # v.ls()
                    # v_next.ls()
                    r.append(v_next - v)
                v = v_next
    if not mark_derived:
        r = r.grib_set_long(["generatingProcessIdentifier", 148])
    return r


def date_from_str(d_str):
    # yyyymmdd
    if len(d_str) == 8:
        return datetime.datetime.strptime(d_str, "%Y%m%d")
    # yyyy-mm-dd ....
    elif len(d_str) >= 10 and d_str[4] == "-" and d_str[7] == "-":
        # yyyy-mm-dd
        if len(d_str) == 10:
            return datetime.datetime.strptime(d_str, "%Y-%m-%d")
        # yyyy-mm-dd hh
        elif len(d_str) == 13:
            return datetime.datetime.strptime(d_str, "%Y-%m-%d %H")
        # yyyy-mm-dd hh:mm
        elif len(d_str) == 16:
            return datetime.datetime.strptime(d_str, "%Y-%m-%d %H:%M")
        # yyyy-mm-dd hh:mm:ss
        elif len(d_str) == 19:
            return datetime.datetime.strptime(d_str, "%Y-%m-%d %H:%M:%S")
    # yyyymmdd.decimal_day
    elif len(d_str) > 8 and d_str[8] == ".":
        f = float("0" + d_str[8:])
        if f >= 1:
            raise ValueError
        else:
            return datetime.datetime.strptime(d_str[:8], "%Y%m%d") + datetime.timedelta(
                seconds=int(f * 86400)
            )
    # mmdd or mdd (as in daily climatologies)
    elif len(d_str) in [3, 4]:
        # try to convert to datatime to see if it is valid date
        d = datetime.datetime.strptime("0004" + d_str.rjust(4, "0"), "%Y%m%d")
        # we just return a tuple since datetime cannot have an invalid date
        return (d.month, d.day)
    # b-dd e.g. apr-02 (as in daily climatologies)
    elif len(d_str) == 6 and d_str[3] == "-":
        months = [
            "jan",
            "feb",
            "mar",
            "apr",
            "may",
            "jun",
            "jul",
            "aug",
            "sep",
            "nov",
            "dec",
        ]
        m = d_str[0:3].lower()
        try:
            m_num = months.index(m) + 1
        except:
            raise ValueError(f"Invalid month={m} specified in date={d_str}!")
        # try to convert to datatime to see if it is valid date
        d = datetime.datetime.strptime("0004" + f"{m_num:02}" + d_str[4:6], "%Y%m%d")
        # we just return a tuple since datetime cannot have an invalid date
        return (d.month, d.day)


def time_from_str(t_str):
    h = m = 0
    if not ":" in t_str:
        # formats: h[mm], hh[mm]
        if len(t_str) in [1, 2]:
            h = int(t_str)
        elif len(t_str) in [3, 4]:
            r = int(t_str)
            h = int(r / 100)
            m = int(r - h * 100)
        else:
            raise Exception(f"Invalid time={t_str}")
    else:
        # formats: h:mm, hh:mm
        lst = t_str.split(":")
        if len(lst) >= 2:
            h = int(lst[0])
            m = int(lst[1])
        else:
            raise Exception(f"Invalid time={t_str}")

    return datetime.time(hour=h, minute=m)


def date_from_ecc_keys(d, t):
    try:
        return datetime.datetime.combine(
            date_from_str(str(d)).date(), time_from_str(str(t))
        )
    except:
        return None


def is_fieldset_type(thing):
    # will return True for binary or python fieldset objects
    return "Fieldset" in thing.__class__.__name__


def get_file_list(path, file_name_pattern=None):
    m = None
    # if isinstance(file_name_pattern, re.Pattern):
    #    m = file_name_pattern.match
    # elif isinstance(file_name_pattern, str):
    if isinstance(file_name_pattern, str):
        if file_name_pattern.startswith('re"'):
            m = re.compile(file_name_pattern[3:-1]).match

    # print(f"path={path} file_name_pattern={file_name_pattern}")

    if m is not None:
        return [os.path.join(path, f) for f in filter(m, os.listdir(path=path))]
    else:
        if isinstance(file_name_pattern, str) and file_name_pattern != "":
            path = os.path.join(path, file_name_pattern)
        if not has_globbing(path):
            return [path]
        else:
            return sorted(glob.glob(path))


def has_globbing(text):
    for x in ["*", "?"]:
        if x in text:
            return True
    if "[" in text and "]" in text:
        return True
    else:
        return False


def unpack(file_path, remove=False):
    if any(file_path.endswith(x) for x in [".tar", ".tar.gz", ".tar.bz2"]):
        target_dir = os.path.dirname(file_path)
        LOG.debug(f"file_path={file_path} target_dir={target_dir}")
        shutil.unpack_archive(file_path, target_dir)
        if remove:
            os.remove(file_path)


def download(url, target):
    from tqdm import tqdm
    import requests

    resp = requests.get(url, stream=True)
    total = int(resp.headers.get("content-length", 0))
    with open(target, "wb") as file, tqdm(
        desc=target,
        total=total,
        unit="iB",
        unit_scale=True,
        unit_divisor=1024,
    ) as bar:
        for data in resp.iter_content(chunk_size=1024):
            size = file.write(data)
            bar.update(size)


def simple_download(url, target):
    import requests

    r = requests.get(url, allow_redirects=True)
    r.raise_for_status()
    open(target, "wb").write(r.content)


def _smooth_core(fs, repeat, m_func, m_arg, **kwargs):
    """
    Performs spatial smoothing on each field in fs with the given callable
    """
    # extract metadata for each field
    meta = fs.grib_get(["gridType", "Ni", "generatingProcessIdentifier"])

    # the resulting fieldset. We cannot use the Fieldset constructor here!
    res = type(fs)()

    # build result in a loop
    for fld, fld_meta in zip(fs, meta):
        # smoothing only works for regular latlon grids
        if fld_meta[0] == "regular_ll":
            ncol = int(fld_meta[1])
            val = fld.values()
            val = np.reshape(val, (-1, ncol))
            for _ in range(repeat):
                val = m_func(val, m_arg, **kwargs)
            r = fld.set_values(val.flatten())
            if fld_meta[2] is not None:
                try:
                    r = r.grib_set_long(
                        ["generatingProcessIdentifier", int(fld_meta[2])]
                    )
                except:
                    pass
            res.append(r)
        else:
            raise ValueError(
                f"Unsupported gridType={fld_meta[0]} in field={i}. Only regular_ll is accepted!"
            )

    return res


def convolve(fs, weight, repeat=1, **kwargs):
    """
    Performs spatial smoothing on each field in fs with a convolution
    """
    from scipy.ndimage.filters import convolve

    m_arg = weight
    return _smooth_core(fs, repeat, convolve, m_arg, **kwargs)


def smooth_n_point(fs, n=9, repeat=1, **kwargs):
    """
    Performs spatial smoothing on each field in fs with an n-point averaging
    """
    from scipy.ndimage.filters import convolve

    if n == 9:
        weights = np.array(
            [[0.0625, 0.125, 0.0625], [0.125, 0.25, 0.125], [0.0625, 0.125, 0.0625]],
            dtype=float,
        )
        m_arg = weights
    elif n == 5:
        weights = np.array(
            [[0.0, 0.125, 0.0], [0.125, 0.5, 0.125], [0.0, 0.125, 0.0]], dtype=float
        )
        m_arg = weights
    else:
        raise ValueError("smooth_n_point: n must be either 5 or 9!")

    return _smooth_core(fs, repeat, convolve, m_arg, **kwargs)


def smooth_gaussian(fs, sigma=1, repeat=1, **kwargs):
    """
    Performs spatial smoothing on each field in fs with a Gaussian filter
    """
    from scipy.ndimage.filters import gaussian_filter

    m_arg = sigma
    return _smooth_core(fs, repeat, gaussian_filter, m_arg, **kwargs)


class Cache:
    ROOT_DIR = os.path.join(tempfile.gettempdir(), f"mpy_ds_{getpass.getuser()}")

    def all_exists(self, items, path):
        for name in items:
            p = os.path.join(path, name)
            # print(f"p={p}")
            if not os.path.exists(p):
                return False
            elif os.path.isdir(p):
                cont_file = os.path.join(path, f".content_{name}")
                if os.path.exists(cont_file):
                    with open(cont_file, "r") as f:
                        try:
                            for item in f.read().split("\n"):
                                if item and not os.path.exists(
                                    os.path.join(path, item)
                                ):
                                    return False
                        except:
                            return False
                else:
                    return False
        return True

    def make_reference(self, items, path):
        for name in items:
            p = os.path.join(path, name)
            if os.path.isdir(p):
                cont_file = os.path.join(path, f".content_{name}")
                with open(cont_file, "w") as f:
                    for item in Path(p).rglob("*"):
                        if item.is_file():
                            f.write(item.relative_to(path).as_posix() + "\n")


CACHE = Cache()



# (C) Copyright 2017- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import copy
import datetime
import os
from black import assert_equivalent

import numpy as np
import pandas as pd


if "METVIEW_PYTHON_ONLY" not in os.environ:
    import metview as mv
else:
    import metview.metviewpy as mv

from metview.metviewpy.param import ParamInfo
from metview.metviewpy.indexer import GribIndexer

PATH = os.path.dirname(__file__)

DB_COLUMNS = copy.deepcopy(GribIndexer.DEFAULT_KEYS)
DB_COLUMNS["_msgIndex1"] = ("l", np.int64, False)
DB_COLUMNS_WIND2 = copy.deepcopy(DB_COLUMNS)
DB_COLUMNS_WIND2["_msgIndex2"] = ("l", np.int64, False)
DB_DEFAULT_COLUMN_NAMES = list(GribIndexer.DEFAULT_KEYS.keys())


def file_in_testdir(filename):
    return os.path.join(PATH, filename)


def build_index_db_dataframe(column_data, key_def=None):
    c = {v: column_data[i] for i, v in enumerate(list(key_def.keys()))}
    pd_types = {k: v[1] for k, v in key_def.items()}
    return pd.DataFrame(c).astype(pd_types)


def file_in_sort_dir(filename):
    return os.path.join(PATH, "sort", filename)


def build_metadata_dataframe(fs, keys):
    val = fs.grib_get(keys, "key")
    md = {k: v for k, v in zip(keys, val)}
    return pd.DataFrame.from_dict(md)


def read_sort_meta_from_csv(name):
    f_name = file_in_sort_dir(f"{name}.csv.gz")
    return pd.read_csv(f_name, index_col=None, dtype=str)


def write_sort_meta_to_csv(name, md):
    f_name = file_in_sort_dir(f"{name}.csv.gz")
    md.to_csv(path_or_buf=f_name, header=True, index=False, compression="gzip")


def test_fieldset_select_single_file():
    f = mv.read(file_in_testdir("tuv_pl.grib"))
    assert f._db is None

    # ------------------------
    # single resulting field
    # ------------------------
    g = f.select(shortName="u", level=700)
    assert len(g) == 1
    assert mv.grib_get(g, ["shortName", "level:l"]) == [["u", 700]]
    g1 = f[7]
    d = g - g1
    assert np.allclose(d.values(), np.zeros(len(d.values())))

    # check index db contents
    assert g._db is not None
    assert "scalar" in g._db.blocks
    assert len(g._db.blocks) == 1
    md = [
        ["u"],
        [131],
        [20180801],
        [1200],
        [0],
        [700],
        ["isobaricInhPa"],
        ["0"],
        ["0001"],
        ["od"],
        ["oper"],
        ["an"],
        [0],
    ]
    df_ref = build_index_db_dataframe(md, key_def=DB_COLUMNS)
    # print(df_ref.dtypes)
    # print(g._db.blocks)
    df = g._db.blocks["scalar"]
    # print(df.dtypes)
    if not df.equals(df_ref):
        print(df.compare(df_ref))
        assert False

    # ------------------------------------
    # single resulting field - paramId
    # ------------------------------------
    g = f.select(paramId=131, level=700)
    assert len(g) == 1
    assert mv.grib_get(g, ["paramId:l", "level:l"]) == [[131, 700]]
    g1 = f[7]
    d = g - g1
    assert np.allclose(d.values(), np.zeros(len(d.values())))

    # check index db contents
    assert g._db is not None
    assert "scalar" in g._db.blocks
    assert len(g._db.blocks) == 1
    md = [
        ["u"],
        [131],
        [20180801],
        [1200],
        [0],
        [700],
        ["isobaricInhPa"],
        ["0"],
        ["0001"],
        ["od"],
        ["oper"],
        ["an"],
        [0],
    ]
    df_ref = build_index_db_dataframe(md, key_def=DB_COLUMNS)
    df = g._db.blocks["scalar"]
    if not df.equals(df_ref):
        print(df.compare(df_ref))
        assert False

    # -------------------------
    # multiple resulting fields
    # -------------------------
    f = mv.read(file_in_testdir("tuv_pl.grib"))
    assert f._db is None

    g = f.select(shortName=["t", "u"], level=[700, 500])
    assert len(g) == 4
    assert mv.grib_get(g, ["shortName", "level:l"]) == [
        ["t", 700],
        ["u", 700],
        ["t", 500],
        ["u", 500],
    ]

    assert g._db is not None
    assert len(g._db.blocks) == 1
    assert "scalar" in g._db.blocks
    md = [
        ["t", "u", "t", "u"],
        [130, 131, 130, 131],
        [20180801] * 4,
        [1200] * 4,
        [0] * 4,
        [700, 700, 500, 500],
        ["isobaricInhPa"] * 4,
        ["0"] * 4,
        ["0001"] * 4,
        ["od"] * 4,
        ["oper"] * 4,
        ["an"] * 4,
        [0, 1, 2, 3],
    ]

    df_ref = build_index_db_dataframe(md, key_def=DB_COLUMNS)
    df = g._db.blocks["scalar"]
    if not df.equals(df_ref):
        print(df.compare(df_ref))
        assert False

    # -------------------------
    # empty result
    # -------------------------
    f = mv.read(file_in_testdir("tuv_pl.grib"))
    g = f.select(shortName="w")
    assert isinstance(g, mv.Fieldset)
    assert len(g) == 0

    # -------------------------
    # invalid key
    # -------------------------
    f = mv.read(file_in_testdir("tuv_pl.grib"))
    g = f.select(INVALIDKEY="w")
    assert isinstance(g, mv.Fieldset)
    assert len(g) == 0

    # -------------------------
    # str or int values
    # -------------------------
    f = mv.read(file_in_testdir("tuv_pl.grib"))
    assert f._db is None

    g = f.select(shortName=["t"], level=["500", 700], marsType="an")
    assert len(g) == 2
    assert mv.grib_get(g, ["shortName", "level:l", "marsType"]) == [
        ["t", 700, "an"],
        ["t", 500, "an"],
    ]

    f = mv.read(file_in_testdir("t_time_series.grib"))
    assert f._db is None

    g = f.select(shortName=["t"], step=[3, 6])
    assert len(g) == 2
    assert mv.grib_get(g, ["shortName", "level:l", "step:l"]) == [
        ["t", 1000, 3],
        ["t", 1000, 6],
    ]

    g = f.select(shortName=["t"], step=["3", "06"])
    assert len(g) == 2
    assert mv.grib_get(g, ["shortName", "level:l", "step:l"]) == [
        ["t", 1000, 3],
        ["t", 1000, 6],
    ]

    # -------------------------
    # repeated use
    # -------------------------
    f = mv.read(file_in_testdir("tuv_pl.grib"))
    assert f._db is None

    g = f.select(shortName=["t"], level=[500, 700], marsType="an")
    assert len(g) == 2
    assert mv.grib_get(g, ["shortName", "level:l", "marsType"]) == [
        ["t", 700, "an"],
        ["t", 500, "an"],
    ]

    g = f.select(shortName=["t"], level=[500], marsType="an")
    assert len(g) == 1
    assert mv.grib_get(g, ["shortName", "level:l", "marsType"]) == [
        ["t", 500, "an"],
    ]

    # -------------------------
    # mars keys
    # -------------------------
    f = mv.read(file_in_testdir("tuv_pl.grib"))
    assert f._db is None

    g = f.select(shortName=["t"], level=[500, 700], marsType="an")
    assert len(g) == 2
    assert mv.grib_get(g, ["shortName", "level:l", "marsType"]) == [
        ["t", 700, "an"],
        ["t", 500, "an"],
    ]

    g = f.select(shortName=["t"], level=[500, 700], type="an")
    assert len(g) == 2
    assert mv.grib_get(g, ["shortName", "level:l", "type"]) == [
        ["t", 700, "an"],
        ["t", 500, "an"],
    ]
    # check the index db contents. "type" must be mapped to the "marsType" column of the
    # db so no rescanning should happen. The db should only contain the default set of columns.
    assert g._db is not None
    assert "scalar" in g._db.blocks
    assert len(g._db.blocks) == 1
    assert list(g._db.blocks["scalar"].keys())[:-1] == DB_DEFAULT_COLUMN_NAMES

    g = f.select(shortName=["t"], level=[500, 700], type="fc")
    assert len(g) == 0

    g = f.select({"shortName": "t", "level": [500, 700], "mars.type": "an"})
    assert len(g) == 2
    assert mv.grib_get(g, ["shortName", "level:l", "mars.type"]) == [
        ["t", 700, "an"],
        ["t", 500, "an"],
    ]

    # -------------------------
    # custom keys
    # -------------------------
    f = mv.read(file_in_testdir("tuv_pl.grib"))
    assert f._db is None

    g = f.select(shortName=["t"], level=[500, 700], gridType="regular_ll")
    assert len(g) == 2
    assert mv.grib_get(g, ["shortName", "level:l", "gridType"]) == [
        ["t", 700, "regular_ll"],
        ["t", 500, "regular_ll"],
    ]

    g = f.select({"shortName": ["t"], "level": [500, 700], "mars.param:s": "130.128"})
    assert len(g) == 2
    assert mv.grib_get(g, ["shortName", "level:l", "mars.param"]) == [
        ["t", 700, "130.128"],
        ["t", 500, "130.128"],
    ]

    assert g._db is not None
    assert "scalar" in g._db.blocks
    assert len(g._db.blocks) == 1
    assert list(g._db.blocks["scalar"].keys())[:-1] == [
        *DB_DEFAULT_COLUMN_NAMES,
        "gridType",
        "mars.param:s",
    ]


def test_fieldset_select_date():
    # date and time
    f = mv.read(file_in_testdir("t_time_series.grib"))
    assert f._db is None

    g = f.select(date="20201221", time="12", step="9")
    assert len(g) == 2

    ref_keys = ["shortName", "date", "time", "step"]
    ref = [
        ["t", "20201221", "1200", "9"],
        ["z", "20201221", "1200", "9"],
    ]

    assert mv.grib_get(g, ref_keys) == ref

    g = f.select(date=20201221, time="1200", step=9)
    assert len(g) == 2
    assert mv.grib_get(g, ref_keys) == ref

    g = f.select(date=20201221, time="12:00", step=9)
    assert len(g) == 2
    assert mv.grib_get(g, ref_keys) == ref

    g = f.select(date=20201221, time=12, step=9)
    assert len(g) == 2
    assert mv.grib_get(g, ref_keys) == ref

    g = f.select(date="2020-12-21", time=1200, step=9)
    assert len(g) == 2
    assert mv.grib_get(g, ref_keys) == ref

    g = f.select(
        date=datetime.datetime(2020, 12, 21),
        time=datetime.time(hour=12, minute=0),
        step=9,
    )
    assert len(g) == 2
    assert mv.grib_get(g, ref_keys) == ref

    # dataDate and dataTime
    g = f.select(dataDate="20201221", dataTime="12", step=9)
    assert len(g) == 2
    assert mv.grib_get(g, ref_keys) == ref

    g = f.select(dataDate="2020-12-21", dataTime="12:00", step=9)
    assert len(g) == 2
    assert mv.grib_get(g, ref_keys) == ref

    # validityDate and validityTime
    g = f.select(validityDate="20201221", validityTime="21")
    assert len(g) == 2
    assert mv.grib_get(g, ref_keys) == ref

    g = f.select(validityDate="2020-12-21", validityTime="21:00")
    assert len(g) == 2
    assert mv.grib_get(g, ref_keys) == ref

    # dateTime
    g = f.select(dateTime="2020-12-21 12:00", step=9)
    assert len(g) == 2
    assert mv.grib_get(g, ref_keys) == ref

    # dataDateTime
    g = f.select(dataDateTime="2020-12-21 12:00", step=9)
    assert len(g) == 2
    assert mv.grib_get(g, ref_keys) == ref

    # validityDateTime
    g = f.select(validityDateTime="2020-12-21 21:00")
    assert len(g) == 2
    assert mv.grib_get(g, ref_keys) == ref

    # ------------------------------------
    # check multiple dates/times
    # ------------------------------------

    ref = [
        ["t", "20201221", "1200", "3"],
        ["z", "20201221", "1200", "3"],
        ["t", "20201221", "1200", "9"],
        ["z", "20201221", "1200", "9"],
    ]

    # date and time
    g = f.select(date="2020-12-21", time=12, step=[3, 9])
    assert len(g) == 4
    assert mv.grib_get(g, ref_keys) == ref

    # dateTime
    g = f.select(dateTime="2020-12-21 12:00", step=[3, 9])
    assert len(g) == 4
    assert mv.grib_get(g, ref_keys) == ref

    # validityDate and validityTime
    g = f.select(validityDate="2020-12-21", validityTime=[15, 21])
    assert len(g) == 4
    assert mv.grib_get(g, ref_keys) == ref

    # validityDateTime
    g = f.select(validityDateTime=["2020-12-21 15:00", "2020-12-21 21:00"])
    assert len(g) == 4
    assert mv.grib_get(g, ref_keys) == ref

    # ------------------------------------
    # check times with 1 digit hours
    # ------------------------------------

    # we create a new fieldset
    f = mv.merge(f[0], mv.grib_set_long(f[2:4], ["time", 600]))

    ref = [
        ["t", "20201221", "0600", "3"],
        ["z", "20201221", "0600", "3"],
    ]

    g = f.select(date="20201221", time="6", step="3")
    assert len(g) == 2
    assert mv.grib_get(g, ref_keys) == ref

    g = f.select(date=20201221, time="06", step=3)
    assert len(g) == 2
    assert mv.grib_get(g, ref_keys) == ref

    g = f.select(date=20201221, time="0600", step=3)
    assert len(g) == 2
    assert mv.grib_get(g, ref_keys) == ref

    g = f.select(date=20201221, time="06:00", step=3)
    assert len(g) == 2
    assert mv.grib_get(g, ref_keys) == ref

    g = f.select(validityDate="2020-12-21", validityTime=9)
    assert len(g) == 2
    assert mv.grib_get(g, ref_keys) == ref

    g = f.select(validityDate="2020-12-21", validityTime="09")
    assert len(g) == 2
    assert mv.grib_get(g, ref_keys) == ref

    g = f.select(validityDate="2020-12-21", validityTime=18)
    assert len(g) == 0

    # ------------------------------------
    # daily climatology dates (no year)
    # ------------------------------------

    f = mv.read(file_in_testdir("daily_clims.grib"))

    g = f.select(date="apr-01")
    assert len(g) == 1
    assert int(mv.grib_get_long(g, "date")) == 401

    g = f.select(date="Apr-02")
    assert len(g) == 1
    assert int(mv.grib_get_long(g, "date")) == 402

    g = f.select(date="402")
    assert len(g) == 1
    assert int(mv.grib_get_long(g, "date")) == 402

    g = f.select(date="0402")
    assert len(g) == 1
    assert int(mv.grib_get_long(g, "date")) == 402

    g = f.select(date=401)
    assert len(g) == 1
    assert int(mv.grib_get_long(g, "date")) == 401

    g = f.select(date=[401, 402])
    assert len(g) == 2
    assert [int(v) for v in mv.grib_get_long(g, "date")] == [402, 401]

    g = f.select(dataDate="apr-01")
    assert len(g) == 1
    assert int(mv.grib_get_long(g, "dataDate")) == 401


def test_fieldset_select_multi_file():
    f = mv.read(file_in_testdir("tuv_pl.grib"))
    f.append(mv.read(file_in_testdir("ml_data.grib")))
    assert f._db is None

    # single resulting field
    g = f.select(shortName="t", level=61)
    # print(f._db.blocks)
    assert len(g) == 1
    assert mv.grib_get(g, ["shortName", "level:l", "typeOfLevel"]) == [
        ["t", 61, "hybrid"]
    ]

    g1 = f[34]
    d = g - g1
    assert np.allclose(d.values(), np.zeros(len(d.values())))

    assert g._db is not None
    assert len(g._db.blocks) == 1
    assert "scalar" in g._db.blocks
    md = [
        ["t"],
        [130],
        [20180111],
        [1200],
        [12],
        [61],
        ["hybrid"],
        [None],
        ["0001"],
        ["od"],
        ["oper"],
        ["fc"],
        [0],
    ]
    df_ref = build_index_db_dataframe(md, key_def=DB_COLUMNS)
    df = g._db.blocks["scalar"]

    if not df.equals(df_ref):
        print(df.compare(df_ref))
        assert False


def test_param_info():
    # no extra info
    p = ParamInfo.build_from_name("2t")
    assert p.name == "2t"
    assert p.scalar == True
    assert p.meta["typeOfLevel"] == "surface"
    assert p.meta["level"] is None

    p = ParamInfo.build_from_name("msl")
    assert p.name == "msl"
    assert p.scalar == True
    assert p.meta["typeOfLevel"] == "surface"
    assert p.meta["level"] is None

    p = ParamInfo.build_from_name("t500")
    assert p.name == "t"
    assert p.scalar == True
    assert p.meta["typeOfLevel"] == "isobaricInhPa"
    assert p.meta["level"] == 500

    p = ParamInfo.build_from_name("t500hPa")
    assert p.name == "t"
    assert p.scalar == True
    assert p.meta["typeOfLevel"] == "isobaricInhPa"
    assert p.meta["level"] == 500

    p = ParamInfo.build_from_name("t")
    assert p.name == "t"
    assert p.scalar == True
    assert p.meta["typeOfLevel"] == ""
    assert p.meta["level"] is None

    p = ParamInfo.build_from_name("t320K")
    assert p.name == "t"
    assert p.scalar == True
    assert p.meta["typeOfLevel"] == "theta"
    assert p.meta["level"] == 320

    p = ParamInfo.build_from_name("t72ml")
    assert p.name == "t"
    assert p.scalar == True
    assert p.meta["typeOfLevel"] == "hybrid"
    assert p.meta["level"] == 72

    p = ParamInfo.build_from_name("wind10m")
    assert p.name == "wind10m"
    assert p.scalar == False
    assert p.meta["typeOfLevel"] == "surface"
    assert p.meta["level"] is None

    p = ParamInfo.build_from_name("wind100")
    assert p.name == "wind"
    assert p.scalar == False
    assert p.meta["typeOfLevel"] == "isobaricInhPa"
    assert p.meta["level"] == 100

    p = ParamInfo.build_from_name("wind700")
    assert p.name == "wind"
    assert p.scalar == False
    assert p.meta["typeOfLevel"] == "isobaricInhPa"
    assert p.meta["level"] == 700

    p = ParamInfo.build_from_name("wind")
    assert p.name == "wind"
    assert p.scalar == False
    assert p.meta["typeOfLevel"] == ""
    assert p.meta["level"] == None

    p = ParamInfo.build_from_name("wind3d")
    assert p.name == "wind3d"
    assert p.scalar == False
    assert p.meta["typeOfLevel"] == ""
    assert p.meta["level"] == None

    p = ParamInfo.build_from_name("wind3d500")
    assert p.name == "wind3d"
    assert p.scalar == False
    assert p.meta["typeOfLevel"] == "isobaricInhPa"
    assert p.meta["level"] == 500

    # exta info
    param_level_types = {
        "2t": ["surface"],
        "msl": ["surface"],
        "wind10m": ["surface"],
        "t": ["isobaricInhPa", "theta"],
        "wind": ["isobaricInhPa"],
    }

    p = ParamInfo.build_from_name("2t", param_level_types=param_level_types)
    assert p.name == "2t"
    assert p.scalar == True
    assert p.meta["typeOfLevel"] == "surface"
    assert p.meta["level"] == None

    try:
        p = ParamInfo.build_from_name("22t", param_level_types=param_level_types)
        assert False
    except:
        pass

    # p = ParamInfo.build_from_name("t2", param_level_types=param_level_types)
    # assert p.name == "2t"
    # assert p.level_type == "surface"
    # assert p.level is None

    p = ParamInfo.build_from_name("msl", param_level_types=param_level_types)
    assert p.name == "msl"
    assert p.scalar == True
    assert p.meta["typeOfLevel"] == "surface"
    assert p.meta["level"] == None

    p = ParamInfo.build_from_name("t500", param_level_types=param_level_types)
    assert p.name == "t"
    assert p.scalar == True
    assert p.meta["typeOfLevel"] == "isobaricInhPa"
    assert p.meta["level"] == 500

    p = ParamInfo.build_from_name("t500hPa", param_level_types=param_level_types)
    assert p.name == "t"
    assert p.scalar == True
    assert p.meta["typeOfLevel"] == "isobaricInhPa"
    assert p.meta["level"] == 500

    p = ParamInfo.build_from_name("t", param_level_types=param_level_types)
    assert p.name == "t"
    assert p.scalar == True
    assert "typeOfLevel" not in p.meta
    assert "level" not in p.meta

    p = ParamInfo.build_from_name("t320K", param_level_types=param_level_types)
    assert p.name == "t"
    assert p.scalar == True
    assert p.meta["typeOfLevel"] == "theta"
    assert p.meta["level"] == 320

    try:
        p = ParamInfo.build_from_name("t72ml", param_level_types=param_level_types)
        assert False
    except:
        pass

    p = ParamInfo.build_from_name("wind10m", param_level_types=param_level_types)
    assert p.name == "wind10m"
    assert p.scalar == False
    assert p.meta["typeOfLevel"] == "surface"
    assert p.meta["level"] is None

    p = ParamInfo.build_from_name("wind100", param_level_types=param_level_types)
    assert p.name == "wind"
    assert p.scalar == False
    assert p.meta["typeOfLevel"] == "isobaricInhPa"
    assert p.meta["level"] == 100

    p = ParamInfo.build_from_name("wind700", param_level_types=param_level_types)
    assert p.name == "wind"
    assert p.scalar == False
    assert p.meta["typeOfLevel"] == "isobaricInhPa"
    assert p.meta["level"] == 700

    p = ParamInfo.build_from_name("wind", param_level_types=param_level_types)
    assert p.name == "wind"
    assert p.scalar == False
    assert p.meta["typeOfLevel"] == "isobaricInhPa"
    assert p.meta["level"] is None

    try:
        p = ParamInfo.build_from_name("wind3d", param_level_types=param_level_types)
        assert False
    except:
        pass

    param_level_types["wind3d"] = ["isobaricInhPa"]
    p = ParamInfo.build_from_name("wind3d", param_level_types=param_level_types)
    assert p.name == "wind3d"
    assert p.scalar == False
    assert p.meta["typeOfLevel"] == "isobaricInhPa"
    assert p.meta["level"] is None

    p = ParamInfo.build_from_name("wind3d500", param_level_types=param_level_types)
    assert p.name == "wind3d"
    assert p.scalar == False
    assert p.meta["typeOfLevel"] == "isobaricInhPa"
    assert p.meta["level"] == 500


def test_param_info_from_fs_single_file():
    if "METVIEW_PYTHON_ONLY" in os.environ:
        return

    f = mv.read(file_in_testdir("tuv_pl.grib"))
    g = f["u700"]
    p = g.ds_param_info
    assert len(g) == 1
    assert p.name == "u"
    assert p.scalar == True
    md = {
        "shortName": "u",
        "paramId": 131,
        "date": 20180801,
        "time": 1200,
        "step": 0,
        "level": 700,
        "typeOfLevel": "isobaricInhPa",
        "number": "0",
        "experimentVersionNumber": "0001",
        "marsClass": "od",
        "marsStream": "oper",
        "marsType": "an",
        "_msgIndex1": 0,
    }
    assert md == p.meta

    g = f["wind500"]
    p = g.ds_param_info
    assert len(g) == 2
    assert p.name == "wind"
    assert p.scalar == False
    md = {
        "shortName": "wind",
        "paramId": 131,
        "date": 20180801,
        "time": 1200,
        "step": 0,
        "level": 500,
        "typeOfLevel": "isobaricInhPa",
        "number": "0",
        "experimentVersionNumber": "0001",
        "marsClass": "od",
        "marsStream": "oper",
        "marsType": "an",
        "_msgIndex1": 0,
        "_msgIndex2": 1,
    }
    assert md == p.meta

    # we lose the db
    g = g + 0
    p = g.ds_param_info
    assert len(g) == 2
    assert p.name == "wind"
    assert p.scalar == False
    md = {
        "shortName": "u",
        "paramId": 131,
        "date": 20180801,
        "time": 1200,
        "step": 0,
        "level": 500,
        "typeOfLevel": "isobaricInhPa",
        "number": "0",
        "experimentVersionNumber": "0001",
        "marsClass": "od",
        "marsStream": "oper",
        "marsType": "an",
    }
    assert md == p.meta

    g = f["t"]
    p = g.ds_param_info
    assert len(g) == 6
    assert p.name == "t"
    assert p.scalar == True
    md = {
        "shortName": "t",
        "paramId": 130,
        "date": 20180801,
        "time": 1200,
        "step": 0,
        "level": None,
        "typeOfLevel": "isobaricInhPa",
        "number": "0",
        "experimentVersionNumber": "0001",
        "marsClass": "od",
        "marsStream": "oper",
        "marsType": "an",
        "_msgIndex1": 0,
    }
    assert md == p.meta

    # we lose the db
    g = g + 0
    p = g.ds_param_info
    assert len(g) == 6
    assert p.name == "t"
    assert p.scalar == True
    md = {
        "shortName": "t",
        "paramId": 130,
        "date": 20180801,
        "time": 1200,
        "step": 0,
        "level": 1000,
        "typeOfLevel": "isobaricInhPa",
        "number": "0",
        "experimentVersionNumber": "0001",
        "marsClass": "od",
        "marsStream": "oper",
        "marsType": "an",
    }
    assert md == p.meta


def test_fieldset_select_operator_single_file():
    f = mv.read(file_in_testdir("tuv_pl.grib"))

    g = f["u700"]
    assert f._db is not None
    assert g._db is not None
    assert len(g) == 1
    assert mv.grib_get(g, ["shortName", "level:l"]) == [["u", 700]]

    g1 = f[7]
    d = g - g1
    assert np.allclose(d.values(), np.zeros(len(d.values())))

    g = f["t"]
    assert len(g) == 6
    assert mv.grib_get(g, ["shortName", "level:l"]) == [
        ["t", 1000],
        ["t", 850],
        ["t", 700],
        ["t", 500],
        ["t", 400],
        ["t", 300],
    ]

    try:
        g = f["w"]
        assert False
    except:
        pass


def test_fieldset_select_operator_multi_file():
    f = mv.read(file_in_testdir("tuv_pl.grib"))
    f.append(mv.read(file_in_testdir("ml_data.grib")))
    assert f._db is None

    # single resulting field
    g = f["t61ml"]
    assert f._db is not None
    assert g._db is not None
    assert len(g) == 1
    assert mv.grib_get(g, ["shortName", "level:l", "typeOfLevel"]) == [
        ["t", 61, "hybrid"]
    ]

    g1 = f[34]
    d = g - g1
    assert np.allclose(d.values(), np.zeros(len(d.values())))


def test_indexer_dataframe_sort_value_with_key():

    md = {
        "paramId": [1, 2, 1, 2, 3],
        "level": [925, 850, 925, 850, 850],
        "step": [12, 110, 1, 3, 1],
        "rest": ["1", "2", "aa", "b1", "1b"],
    }

    md_ref = {
        "paramId": [1, 1, 2, 2, 3],
        "level": [925, 925, 850, 850, 850],
        "step": [1, 12, 3, 110, 1],
        "rest": ["aa", "1", "b1", "2", "1b"],
    }

    df = pd.DataFrame(md)
    df = GribIndexer._sort_dataframe(df)
    df_ref = pd.DataFrame(md_ref)

    if not df.equals(df_ref):
        print(df.compare(df_ref))
        assert False


def test_describe():

    f = mv.read(file_in_testdir("tuv_pl.grib"))

    # full contents
    df = f.describe(no_print=True)

    ref = {
        "typeOfLevel": {
            "t": "isobaricInhPa",
            "u": "isobaricInhPa",
            "v": "isobaricInhPa",
        },
        "level": {"t": "300,400,...", "u": "300,400,...", "v": "300,400,..."},
        "date": {"t": 20180801, "u": 20180801, "v": 20180801},
        "time": {"t": 1200, "u": 1200, "v": 1200},
        "step": {"t": 0, "u": 0, "v": 0},
        "paramId": {"t": 130, "u": 131, "v": 132},
        "class": {"t": "od", "u": "od", "v": "od"},
        "stream": {"t": "oper", "u": "oper", "v": "oper"},
        "type": {"t": "an", "u": "an", "v": "an"},
        "experimentVersionNumber": {"t": "0001", "u": "0001", "v": "0001"},
    }
    ref_full = ref

    assert ref == df.to_dict()

    df = f.describe()
    assert ref == df.to_dict()

    # single param by shortName
    df = f.describe("t", no_print=True)

    ref = {
        "val": {
            "shortName": "t",
            "name": "Temperature",
            "paramId": 130,
            "units": "K",
            "typeOfLevel": "isobaricInhPa",
            "level": "300,400,500,700,850,1000",
            "date": "20180801",
            "time": "1200",
            "step": "0",
            "class": "od",
            "stream": "oper",
            "type": "an",
            "experimentVersionNumber": "0001",
        }
    }

    assert ref["val"] == df["val"].to_dict()

    df = f.describe(param="t", no_print=True)
    assert ref == df.to_dict()
    df = f.describe("t")
    assert ref == df.to_dict()
    df = f.describe(param="t")
    assert ref == df.to_dict()

    # single param by paramId
    df = f.describe(130, no_print=True)

    ref = {
        "val": {
            "shortName": "t",
            "name": "Temperature",
            "paramId": 130,
            "units": "K",
            "typeOfLevel": "isobaricInhPa",
            "level": "300,400,500,700,850,1000",
            "date": "20180801",
            "time": "1200",
            "step": "0",
            "class": "od",
            "stream": "oper",
            "type": "an",
            "experimentVersionNumber": "0001",
        }
    }

    assert ref == df.to_dict()

    df = f.describe(param=130, no_print=True)
    assert ref == df.to_dict()
    df = f.describe(130)
    assert ref == df.to_dict()
    df = f.describe(param=130)
    assert ref == df.to_dict()

    # append
    g = f + 0
    df = g.describe(no_print=True)
    assert ref_full == df.to_dict()

    g.append(f[0].grib_set_long(["level", 25]))
    df = g.describe(no_print=True)

    ref = {
        "typeOfLevel": {
            "t": "isobaricInhPa",
            "u": "isobaricInhPa",
            "v": "isobaricInhPa",
        },
        "level": {"t": "25,300,...", "u": "300,400,...", "v": "300,400,..."},
        "date": {"t": 20180801, "u": 20180801, "v": 20180801},
        "time": {"t": 1200, "u": 1200, "v": 1200},
        "step": {"t": 0, "u": 0, "v": 0},
        "paramId": {"t": 130, "u": 131, "v": 132},
        "class": {"t": "od", "u": "od", "v": "od"},
        "stream": {"t": "oper", "u": "oper", "v": "oper"},
        "type": {"t": "an", "u": "an", "v": "an"},
        "experimentVersionNumber": {"t": "0001", "u": "0001", "v": "0001"},
    }

    assert ref == df.to_dict()


def test_ls():

    f = mv.read(file_in_testdir("tuv_pl.grib"))

    # default keys
    df = f[:4].ls(no_print=True)

    ref = {
        "centre": {0: "ecmf", 1: "ecmf", 2: "ecmf", 3: "ecmf"},
        "shortName": {0: "t", 1: "u", 2: "v", 3: "t"},
        "typeOfLevel": {
            0: "isobaricInhPa",
            1: "isobaricInhPa",
            2: "isobaricInhPa",
            3: "isobaricInhPa",
        },
        "level": {0: 1000, 1: 1000, 2: 1000, 3: 850},
        "dataDate": {0: 20180801, 1: 20180801, 2: 20180801, 3: 20180801},
        "dataTime": {0: 1200, 1: 1200, 2: 1200, 3: 1200},
        "stepRange": {0: "0", 1: "0", 2: "0", 3: "0"},
        "dataType": {0: "an", 1: "an", 2: "an", 3: "an"},
        "gridType": {
            0: "regular_ll",
            1: "regular_ll",
            2: "regular_ll",
            3: "regular_ll",
        },
    }

    assert ref == df.to_dict()

    # extra keys
    df = f[:2].ls(extra_keys=["paramId"], no_print=True)

    ref = {
        "centre": {0: "ecmf", 1: "ecmf"},
        "shortName": {0: "t", 1: "u"},
        "typeOfLevel": {0: "isobaricInhPa", 1: "isobaricInhPa"},
        "level": {0: 1000, 1: 1000},
        "dataDate": {0: 20180801, 1: 20180801},
        "dataTime": {0: 1200, 1: 1200},
        "stepRange": {0: "0", 1: "0"},
        "dataType": {0: "an", 1: "an"},
        "gridType": {0: "regular_ll", 1: "regular_ll"},
        "paramId": {0: 130, 1: 131},
    }

    assert ref == df.to_dict()

    # filter
    df = f.ls(filter={"shortName": ["t", "v"], "level": 850}, no_print=True)

    ref = {
        "centre": {3: "ecmf", 5: "ecmf"},
        "shortName": {3: "t", 5: "v"},
        "typeOfLevel": {3: "isobaricInhPa", 5: "isobaricInhPa"},
        "level": {3: 850, 5: 850},
        "dataDate": {3: 20180801, 5: 20180801},
        "dataTime": {3: 1200, 5: 1200},
        "stepRange": {3: "0", 5: "0"},
        "dataType": {3: "an", 5: "an"},
        "gridType": {3: "regular_ll", 5: "regular_ll"},
    }

    assert ref == df.to_dict()

    # append
    g = f[:2]
    df = g.ls(no_print=True)

    ref = {
        "centre": {0: "ecmf", 1: "ecmf"},
        "shortName": {0: "t", 1: "u"},
        "typeOfLevel": {
            0: "isobaricInhPa",
            1: "isobaricInhPa",
        },
        "level": {0: 1000, 1: 1000},
        "dataDate": {0: 20180801, 1: 20180801},
        "dataTime": {0: 1200, 1: 1200},
        "stepRange": {0: "0", 1: "0"},
        "dataType": {0: "an", 1: "an"},
        "gridType": {
            0: "regular_ll",
            1: "regular_ll",
        },
    }

    assert ref == df.to_dict()

    g.append(f[2].grib_set_long(["level", 500]))
    df = g.ls(no_print=True)
    ref = {
        "centre": {0: "ecmf", 1: "ecmf", 2: "ecmf"},
        "shortName": {0: "t", 1: "u", 2: "v"},
        "typeOfLevel": {
            0: "isobaricInhPa",
            1: "isobaricInhPa",
            2: "isobaricInhPa",
        },
        "level": {0: 1000, 1: 1000, 2: 500},
        "dataDate": {0: 20180801, 1: 20180801, 2: 20180801},
        "dataTime": {0: 1200, 1: 1200, 2: 1200},
        "stepRange": {0: "0", 1: "0", 2: "0"},
        "dataType": {0: "an", 1: "an", 2: "an"},
        "gridType": {
            0: "regular_ll",
            1: "regular_ll",
            2: "regular_ll",
        },
    }

    assert ref == df.to_dict()


def test_sort():

    # In each message the message index (1 based) is encoded
    # into latitudeOfLastGridPoint!
    fs = mv.read(file_in_sort_dir("sort_data.grib"))

    default_sort_keys = ["date", "time", "step", "number", "level", "paramId"]
    assert GribIndexer.DEFAULT_SORT_KEYS == default_sort_keys

    # Note: shortName, units and latitudeOfLastGridPoint are non-default sort keys!
    keys = [
        "date",
        "time",
        "step",
        "number",
        "level",
        "paramId",
        "shortName",
        "units",
        "latitudeOfLastGridPoint",
    ]

    # the reference csv files were generated like this:
    #   write_sort_meta_to_csv(f_name, md)
    #
    # the correctness of the reference was tested by generating md_ref using
    # the GribIndexer and comparing it to md. E.g.:
    #   md_ori = build_metadata_dataframe(fs, keys)
    #   md_ref = GribIndexer._sort_dataframe(md_ori, columns=default_sort_keys)
    #   assert md.equals(md_ref)

    # default sorting
    f_name = "default"
    r = fs.sort()
    assert len(fs) == len(r)
    md = build_metadata_dataframe(r, keys)
    md_ref = read_sort_meta_from_csv(f_name)
    if not md.equals(md_ref):
        print(md.compare(md_ref))
        assert False

    # -----------------------------
    # default sorting direction
    # -----------------------------

    sort_keys = {
        k: k
        for k in [
            "date",
            "time",
            "step",
            "number",
            "level",
            "paramId",
            "shortName",
            "units",
        ]
    }
    sort_keys["date_level"] = ["date", "level"]
    sort_keys["level_units"] = ["level", "units"]
    sort_keys["keys"] = keys

    for f_name, key in sort_keys.items():
        r = fs.sort(key)
        assert len(fs) == len(r)
        md = build_metadata_dataframe(r, keys)
        md_ref = read_sort_meta_from_csv(f_name)
        if not md.equals(md_ref):
            print(md.compare(md_ref))
            assert False, f"key={key}"

    # single key as list
    key = f_name = "level"
    r = fs.sort([key])
    assert len(fs) == len(r)
    md = build_metadata_dataframe(r, keys)
    md_ref = read_sort_meta_from_csv(f_name)
    if not md.equals(md_ref):
        print(md.compare(md_ref))
        assert False, f"key={key}"

    # -----------------------------
    # custom sorting direction
    # -----------------------------

    # default keys

    # ascending
    f_name = "default"
    r = fs.sort(ascending=True)
    assert len(fs) == len(r)
    md = build_metadata_dataframe(r, keys)
    md_ref = read_sort_meta_from_csv(f_name)
    if not md.equals(md_ref):
        print(md.compare(md_ref))
        assert False, "default ascending"

    # descending
    f_name = "default_desc"
    r = fs.sort(ascending=False)
    assert len(fs) == len(r)
    md = build_metadata_dataframe(r, keys)
    md_ref = read_sort_meta_from_csv(f_name)
    if not md.equals(md_ref):
        print(md.compare(md_ref))
        assert False, "default descending"

    # single key
    key = "level"

    # ascending
    f_name = f"{key}_asc"
    r = fs.sort(key, "<")
    assert len(fs) == len(r)
    md = build_metadata_dataframe(r, keys)
    md_ref = read_sort_meta_from_csv(f_name)
    if not md.equals(md_ref):
        print(md.compare(md_ref))
        assert False, f"key={key}"

    r = fs.sort(key, ["<"])
    assert len(fs) == len(r)
    md = build_metadata_dataframe(r, keys)
    if not md.equals(md_ref):
        print(md.compare(md_ref))
        assert False, f"key={key}"

    r = fs.sort(key, ascending=True)
    assert len(fs) == len(r)
    md = build_metadata_dataframe(r, keys)
    if not md.equals(md_ref):
        print(md.compare(md_ref))
        assert False, f"key={key}"

    # descending
    f_name = f"{key}_desc"
    r = fs.sort(key, ">")
    assert len(fs) == len(r)
    md = build_metadata_dataframe(r, keys)
    md_ref = read_sort_meta_from_csv(f_name)
    if not md.equals(md_ref):
        print(md.compare(md_ref))
        assert False, f"key={key}"

    r = fs.sort(key, [">"])
    assert len(fs) == len(r)
    md = build_metadata_dataframe(r, keys)
    if not md.equals(md_ref):
        print(md.compare(md_ref))
        assert False, f"key={key}"

    r = fs.sort(key, ascending=False)
    assert len(fs) == len(r)
    md = build_metadata_dataframe(r, keys)
    if not md.equals(md_ref):
        print(md.compare(md_ref))
        assert False, f"key={key}"

    # multiple keys
    key = ["level", "paramId", "date"]

    f_name = "multi_asc"
    r = fs.sort(key, "<")
    assert len(fs) == len(r)
    md = build_metadata_dataframe(r, keys)
    md_ref = read_sort_meta_from_csv(f_name)
    if not md.equals(md_ref):
        print(md.compare(md_ref))
        assert False, f"key={key}"

    r = fs.sort(key, ascending=True)
    assert len(fs) == len(r)
    md = build_metadata_dataframe(r, keys)
    if not md.equals(md_ref):
        print(md.compare(md_ref))
        assert False, f"key={key}"

    f_name = "multi_desc"
    r = fs.sort(key, ">")
    assert len(fs) == len(r)
    md = build_metadata_dataframe(r, keys)
    md_ref = read_sort_meta_from_csv(f_name)
    if not md.equals(md_ref):
        print(md.compare(md_ref))
        assert False, f"key={key}"

    r = fs.sort(key, ascending=False)
    assert len(fs) == len(r)
    md = build_metadata_dataframe(r, keys)
    if not md.equals(md_ref):
        print(md.compare(md_ref))
        assert False, f"key={key}"

    f_name = "multi_mixed"
    r = fs.sort(key, ["<", ">", "<"])
    assert len(fs) == len(r)
    md = build_metadata_dataframe(r, keys)
    md_ref = read_sort_meta_from_csv(f_name)
    if not md.equals(md_ref):
        print(md.compare(md_ref))
        assert False, f"key={key}"

    r = fs.sort(key, ascending=[True, False, True])
    assert len(fs) == len(r)
    md = build_metadata_dataframe(r, keys)
    if not md.equals(md_ref):
        print(md.compare(md_ref))
        assert False, f"key={key}"

    # invalid arguments
    try:
        r = fs.sort(key, [">", "<"])
        assert False
    except ValueError:
        pass

    try:
        r = fs.sort(key, "1")
        assert False
    except ValueError:
        pass

    try:
        r = fs.sort(key, ascending=["True", "False"])
        assert False
    except ValueError:
        pass

    try:
        r = fs.sort(key, ">", ascending="True")
        assert False
    except ValueError:
        pass

    try:
        r = fs.sort(key, ">", ascending=["True", "False"])
        assert False
    except ValueError:
        pass


def test_speed():
    # test with grib written with write() function
    fs = mv.Fieldset(path=os.path.join(PATH, "tuv_pl.grib"))

    fs_u = fs.select(shortName="u")
    fs_v = fs.select(shortName="v")

    # single field
    u = fs_u[0]
    v = fs_v[0]
    r = mv.speed(u, v)
    assert len(r) == 1
    np.testing.assert_allclose(
        r.values(), np.sqrt(np.square(u.values()) + np.square(v.values())), rtol=1e-05
    )
    assert r.grib_get_long("paramId") == 10

    # multi fields
    u = fs_u[:2]
    v = fs_v[:2]
    r = mv.speed(u, v)
    assert len(r) == 2
    for i in range(len(r)):
        np.testing.assert_allclose(
            r[i].values(),
            np.sqrt(np.square(u[i].values()) + np.square(v[i].values())),
            rtol=1e-05,
        )
    assert r.grib_get_long("paramId") == [10, 10]

    # no arguments
    u = fs_u
    v = fs_v
    r = fs["wind"].speed()
    assert len(r) == 6
    for i in range(len(r)):
        np.testing.assert_allclose(
            r[i].values(),
            np.sqrt(np.square(u[i].values()) + np.square(v[i].values())),
            rtol=1e-05,
        )
    assert r.grib_get_long("paramId") == [10] * 6


def test_deacc():
    f = mv.Fieldset(path=os.path.join(PATH, "t_time_series.grib"))[:3]

    r = f.deacc()
    assert len(r) == len(f)
    assert r.grib_get_long("generatingProcessIdentifier") == [148] * len(r)
    for i in range(len(f)):
        v_ref = f[0].values() * 0 if i == 0 else f[i].values() - f[i - 1].values()
        np.testing.assert_allclose(r[i].values(), v_ref, rtol=1e-03)

    r = f.deacc(key="")
    assert len(r) == len(f)
    assert r.grib_get_long("generatingProcessIdentifier") == [148] * len(r)
    for i in range(len(f)):
        v_ref = f[0].values() * 0 if i == 0 else f[i].values() - f[i - 1].values()
        np.testing.assert_allclose(r[i].values(), v_ref, rtol=1e-03)

    r = f.deacc(skip_first=True)
    assert len(r) == len(f) - 1
    assert r.grib_get_long("generatingProcessIdentifier") == [148] * len(r)
    for i in range(len(r)):
        v_ref = f[i + 1].values() - f[i].values()
        np.testing.assert_allclose(r[i].values(), v_ref, rtol=1e-03)

    r = f.deacc(skip_first=True, mark_derived=True)
    assert len(r) == len(f) - 1
    # assert r.grib_get_long("generatingProcessIdentifier") == [254] * len(r)
    for i in range(len(r)):
        v_ref = f[i + 1].values() - f[i].values()
        np.testing.assert_allclose(r[i].values(), v_ref, rtol=1e-03)

    # use grouping by key
    f = mv.Fieldset(path=os.path.join(PATH, "t_time_series.grib"))[:6]

    # only "step" is part of the default set of indexing keys!
    keys = ["step", "startStep", "stepRange"]
    for key in keys:
        r = f.deacc(key=key)
        assert len(r) == 6
        assert r.grib_get_long("generatingProcessIdentifier") == [148] * len(r)

        v_ref = f[0].values() * 0
        np.testing.assert_allclose(r[0].values(), v_ref, rtol=1e-03)
        v_ref = f[1].values() * 0
        np.testing.assert_allclose(r[1].values(), v_ref, rtol=1e-03)

        steps = {2: (2, 0), 3: (3, 1), 4: (4, 2), 5: (5, 3)}

        for idx, steps in steps.items():
            v_ref = f[steps[0]].values() - f[steps[1]].values()
            np.testing.assert_allclose(r[idx].values(), v_ref, rtol=1e-03)



# (C) Copyright 2017- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from inspect import ArgInfo
import numpy as np
import os
import pytest

import metview.metviewpy as mv
from metview.metviewpy import utils
from metview.metviewpy.temporary import is_temp_file

PATH = os.path.dirname(__file__)


def test_empty_fieldset_contructor():
    f = mv.Fieldset()
    assert type(f) is mv.Fieldset
    assert len(f) == 0


def test_fieldset_contructor_bad_file_path():
    with pytest.raises(FileNotFoundError):
        f = mv.Fieldset(path="does/not/exist")


def test_non_empty_fieldset_contructor_len():
    f = mv.Fieldset(path=os.path.join(PATH, "test.grib"))
    assert type(f) is mv.Fieldset
    assert len(f) == 1


def test_non_empty_fieldset_contructor_len_18():
    f = mv.Fieldset(path=os.path.join(PATH, "tuv_pl.grib"))
    assert type(f) is mv.Fieldset
    assert len(f) == 18


def test_fieldset_create_from_list_of_paths():
    paths = [os.path.join(PATH, "t_for_xs.grib"), os.path.join(PATH, "ml_data.grib")]
    f = mv.Fieldset(path=paths)
    assert len(f) == 42
    assert f[0:2].grib_get_long("level") == [1000, 850]
    assert f[5:9].grib_get_long("level") == [300, 1, 1, 5]
    assert f[40:42].grib_get_long("level") == [133, 137]


def test_fieldset_create_from_glob_path_single():
    f = mv.Fieldset(path=os.path.join(PATH, "test.g*ib"))
    assert type(f) == mv.Fieldset
    assert len(f) == 1


def test_fieldset_create_from_glob_path_multi():
    f = mv.Fieldset(path=os.path.join(PATH, "t_*.grib"))
    assert type(f) == mv.Fieldset
    assert len(f) == 17
    par_ref = [
        ["t", "1000"],
        ["t", "850"],
        ["t", "700"],
        ["t", "500"],
        ["t", "400"],
        ["t", "300"],
        ["t", "1000"],
        ["z", "1000"],
        ["t", "1000"],
        ["z", "1000"],
        ["t", "1000"],
        ["z", "1000"],
        ["t", "1000"],
        ["z", "1000"],
        ["t", "1000"],
        ["z", "1000"],
        ["t", "1000"],
    ]
    assert par_ref == f.grib_get(["shortName", "level"])


def test_fieldset_create_from_glob_paths():
    f = mv.Fieldset(
        path=[os.path.join(PATH, "test.g*ib"), os.path.join(PATH, "t_*.grib")]
    )
    assert type(f) == mv.Fieldset
    assert len(f) == 18
    par_ref = [
        ["2t", "0"],
        ["t", "1000"],
        ["t", "850"],
        ["t", "700"],
        ["t", "500"],
        ["t", "400"],
        ["t", "300"],
        ["t", "1000"],
        ["z", "1000"],
        ["t", "1000"],
        ["z", "1000"],
        ["t", "1000"],
        ["z", "1000"],
        ["t", "1000"],
        ["z", "1000"],
        ["t", "1000"],
        ["z", "1000"],
        ["t", "1000"],
    ]
    assert par_ref == f.grib_get(["shortName", "level"])


def test_read_1():
    f = mv.read(os.path.join(PATH, "test.grib"))
    assert type(f) is mv.Fieldset
    assert len(f) == 1


def test_grib_get_string_1():
    f = mv.Fieldset(path=os.path.join(PATH, "test.grib"))
    sn = f.grib_get_string("shortName")
    assert sn == "2t"


def test_grib_get_string_18():
    f = mv.Fieldset(path=os.path.join(PATH, "tuv_pl.grib"))
    sn = f.grib_get_string("shortName")
    assert sn == ["t", "u", "v"] * 6


def test_grib_get_long_1():
    f = mv.Fieldset(path=os.path.join(PATH, "test.grib"))
    lev = f.grib_get_long("level")
    assert lev == 0


def test_grib_get_long_18():
    f = mv.Fieldset(path=os.path.join(PATH, "tuv_pl.grib"))
    lev = f.grib_get_long("level")
    assert lev == ([1000] * 3) + ([850] * 3) + ([700] * 3) + ([500] * 3) + (
        [400] * 3
    ) + ([300] * 3)


def test_grib_get_double_1():
    f = mv.Fieldset(path=os.path.join(PATH, "test.grib"))
    m = f.grib_get_double("max")
    assert np.isclose(m, 316.061)


def test_grib_get_double_18():
    f = mv.Fieldset(path=os.path.join(PATH, "tuv_pl.grib"))
    m = f.grib_get_double("max")
    ref_m = [
        320.564,
        21.7131,
        19.8335,
        304.539,
        43.1016,
        28.661,
        295.265,
        44.1455,
        31.6385,
        275.843,
        52.74,
        47.0099,
        264.003,
        62.2138,
        55.9496,
        250.653,
        66.4555,
        68.9203,
    ]
    np.testing.assert_allclose(m, ref_m, 0.001)


def test_grib_get_long_array_1():
    f = mv.Fieldset(path=os.path.join(PATH, "rgg_small_subarea_cellarea_ref.grib"))
    pl = f.grib_get_long_array("pl")
    assert isinstance(pl, np.ndarray)
    assert len(pl) == 73
    assert pl[0] == 24
    assert pl[1] == 28
    assert pl[20] == 104
    assert pl[72] == 312


def test_grib_get_double_array_1():
    f = mv.Fieldset(path=os.path.join(PATH, "test.grib"))
    v = f.grib_get_double_array("values")
    assert isinstance(v, np.ndarray)
    assert len(v) == 115680
    assert np.isclose(v[0], 260.4356)
    assert np.isclose(v[24226], 276.1856)
    assert np.isclose(v[36169], 287.9356)
    assert np.isclose(v[115679], 227.1856)


def test_grib_get_double_array_18():
    f = mv.Fieldset(path=os.path.join(PATH, "tuv_pl.grib"))
    v = f.grib_get_double_array("values")
    assert isinstance(v, list)
    assert len(v) == 18
    assert isinstance(v[0], np.ndarray)
    assert isinstance(v[17], np.ndarray)
    assert len(v[0]) == 2664
    assert len(v[17]) == 2664
    eps = 0.001
    assert np.isclose(v[0][0], 272.5642, eps)
    assert np.isclose(v[0][1088], 304.5642, eps)
    assert np.isclose(v[17][0], -3.0797, eps)
    assert np.isclose(v[17][2663], -11.0797, eps)


def test_grib_get_generic():
    f = mv.Fieldset(path=os.path.join(PATH, "tuv_pl.grib"))[0:4]
    sn = f.grib_get(["shortName"])
    assert sn == [["t"], ["u"], ["v"], ["t"]]
    cs = f.grib_get(["centre:s"])
    assert cs == [["ecmf"], ["ecmf"], ["ecmf"], ["ecmf"]]
    cl = f.grib_get(["centre:l"])
    assert cl == [[98], [98], [98], [98]]
    lg = f.grib_get(["level:d", "cfVarName"])
    assert lg == [[1000, "t"], [1000, "u"], [1000, "v"], [850, "t"]]
    lgk = f.grib_get(["level:d", "cfVarName"], "key")
    assert lgk == [[1000, 1000, 1000, 850], ["t", "u", "v", "t"]]
    with pytest.raises(ValueError):
        lgk = f.grib_get(["level:d", "cfVarName"], "silly")
    ln = f.grib_get(["level:n"])
    assert ln == [[1000], [1000], [1000], [850]]
    cn = f.grib_get(["centre:n"])
    assert cn == [["ecmf"], ["ecmf"], ["ecmf"], ["ecmf"]]
    vn = f[0].grib_get(["longitudes:n"])
    assert vn[0][0][0] == 0
    assert vn[0][0][1] == 5
    assert vn[0][0][5] == 25


def test_grib_get_generic_key_not_exist():
    a = mv.Fieldset(path=os.path.join(PATH, "tuv_pl.grib"))[0:2]
    kv = a.grib_get(["silly"])
    assert kv == [[None], [None]]
    with pytest.raises(Exception):
        kv = a.grib_get_long(["silly"])


def test_values_1():
    f = mv.Fieldset(path=os.path.join(PATH, "test.grib"))
    v = f.values()
    assert isinstance(v, np.ndarray)
    assert len(v) == 115680
    assert np.isclose(v[0], 260.4356)
    assert np.isclose(v[24226], 276.1856)
    assert np.isclose(v[36169], 287.9356)
    assert np.isclose(v[115679], 227.1856)


def test_values_18():
    f = mv.Fieldset(path=os.path.join(PATH, "tuv_pl.grib"))
    v = f.values()
    assert isinstance(v, np.ndarray)
    assert v.shape == (18, 2664)
    assert isinstance(v[0], np.ndarray)
    assert isinstance(v[17], np.ndarray)
    assert len(v[0]) == 2664
    assert len(v[17]) == 2664
    eps = 0.001
    assert np.isclose(v[0][0], 272.5642, eps)
    assert np.isclose(v[0][1088], 304.5642, eps)
    assert np.isclose(v[17][0], -3.0797, eps)
    assert np.isclose(v[17][2663], -11.0797, eps)


def test_values_with_missing():
    f = mv.Fieldset(path=os.path.join(PATH, "t_with_missing.grib"))
    v = f.values()
    assert isinstance(v, np.ndarray)
    assert v.shape == (2664,)
    eps = 0.001
    assert np.isclose(v[0], 272.5642, eps)
    assert np.isnan(v[798])
    assert np.isnan(v[806])
    assert np.isnan(v[1447])
    assert np.isclose(v[2663], 240.5642, eps)


def test_grib_set_string():
    f = mv.Fieldset(path=os.path.join(PATH, "tuv_pl.grib"))[0:2]
    g = f.grib_set_string(["pressureUnits", "silly"])
    assert g.grib_get_string("pressureUnits") == ["silly"] * 2
    assert f.grib_get_string("pressureUnits") == ["hPa"] * 2
    g = f.grib_set_string(["pressureUnits", "silly", "shortName", "q"])
    assert g.grib_get_string("pressureUnits") == ["silly"] * 2
    assert g.grib_get_string("shortName") == ["q", "q"]
    assert f.grib_get_string("pressureUnits") == ["hPa"] * 2
    assert f.grib_get_string("shortName") == ["t", "u"]


def test_grib_set_long():
    f = mv.Fieldset(path=os.path.join(PATH, "tuv_pl.grib"))[0:2]
    g = f.grib_set_long(["level", 95])
    assert g.grib_get_long("level") == [95] * 2
    assert f.grib_get_long("level") == [1000] * 2
    g = f.grib_set_long(["level", 95, "time", 1800])
    assert g.grib_get_long("level") == [95] * 2
    assert g.grib_get_long("time") == [1800] * 2
    assert f.grib_get_long("level") == [1000] * 2
    assert f.grib_get_long("time") == [1200] * 2


def test_grib_set_double():
    f = mv.Fieldset(path=os.path.join(PATH, "tuv_pl.grib"))[0:2]
    g = f.grib_set_double(["level", 95])
    assert g.grib_get_double("level") == [95] * 2
    assert f.grib_get_double("level") == [1000] * 2
    orig_point = f.grib_get_double("longitudeOfFirstGridPointInDegrees")
    g = f.grib_set_double(["longitudeOfFirstGridPointInDegrees", 95.6])
    assert g.grib_get_double("longitudeOfFirstGridPointInDegrees") == [95.6] * 2
    assert f.grib_get_double("longitudeOfFirstGridPointInDegrees") == orig_point


def test_grib_set_generic():
    f = mv.Fieldset(path=os.path.join(PATH, "tuv_pl.grib"))[0:2]
    g = f.grib_set(["shortName", "r"])
    assert g.grib_get_string("shortName") == ["r"] * 2
    assert f.grib_get_string("shortName") == ["t", "u"]
    g = f.grib_set(["shortName:s", "q"])
    assert g.grib_get_string("shortName") == ["q"] * 2
    assert f.grib_get_string("shortName") == ["t", "u"]

    g = f.grib_set(["level:l", 500, "shortName", "z"])
    assert g.grib_get_long("level") == [500] * 2
    assert g.grib_get_string("shortName") == ["z"] * 2
    assert f.grib_get_long("level") == [1000] * 2
    assert f.grib_get_string("shortName") == ["t", "u"]

    g = f.grib_set(["level:d", 500])
    np.testing.assert_allclose(
        np.array(g.grib_get_double("level")), np.array([500] * 2)
    )
    np.testing.assert_allclose(
        np.array(f.grib_get_double("level")), np.array([1000] * 2)
    )

    g = f.grib_set_double(["longitudeOfFirstGridPointInDegrees", 95.6])
    np.testing.assert_allclose(
        np.array(g.grib_get_double("longitudeOfFirstGridPointInDegrees")),
        np.array([95.6] * 2),
    )
    np.testing.assert_allclose(
        np.array(f.grib_get_double("longitudeOfFirstGridPointInDegrees")), [0, 0]
    )


def test_write_fieldset():
    f = mv.Fieldset(path=os.path.join(PATH, "tuv_pl.grib"))
    temp_path = "written_tuv_pl.grib"
    f.write(temp_path)
    assert os.path.isfile(temp_path)
    g = mv.Fieldset(path=temp_path)
    assert type(g) == mv.Fieldset
    assert len(g) == 18
    sn = g.grib_get_string("shortName")
    assert sn == ["t", "u", "v"] * 6
    f = 0
    os.remove(temp_path)


def test_write_modified_fieldset_binop():
    f = mv.Fieldset(path=os.path.join(PATH, "tuv_pl.grib"))
    fp20 = f + 20
    temp_path = "written_tuv_pl_binop.grib"
    fp20.write(temp_path)
    assert os.path.isfile(temp_path)
    g = mv.Fieldset(path=temp_path)
    assert type(g) == mv.Fieldset
    assert len(g) == 18
    sn = g.grib_get_string("shortName")
    assert sn == ["t", "u", "v"] * 6
    np.testing.assert_allclose(g.values(), f.values() + 20)
    f = 0
    os.remove(temp_path)


def test_write_modified_fieldset_unop():
    f = mv.Fieldset(path=os.path.join(PATH, "tuv_pl.grib"))
    negf = -f
    temp_path = "written_tuv_pl_unop.grib"
    negf.write(temp_path)
    assert os.path.isfile(temp_path)
    g = mv.Fieldset(path=temp_path)
    assert type(g) == mv.Fieldset
    assert len(g) == 18
    sn = g.grib_get_string("shortName")
    assert sn == ["t", "u", "v"] * 6
    np.testing.assert_allclose(g.values(), -f.values(), 0.0001)
    f = 0
    os.remove(temp_path)


def test_field_func():
    def sqr_func(x):
        return x * x

    f = mv.Fieldset(path=os.path.join(PATH, "tuv_pl.grib"))
    g = f.field_func(sqr_func)
    assert type(g) == mv.Fieldset
    assert len(g) == 18
    vf = f.values()
    vg = g.values()
    np.testing.assert_allclose(vg, vf * vf, 0.0001)


def test_field_func_neg():
    f = mv.Fieldset(path=os.path.join(PATH, "tuv_pl.grib"))
    g = -f
    assert type(g) == mv.Fieldset
    assert len(g) == 18
    vf = f.values()
    vg = g.values()
    np.testing.assert_allclose(vg, -vf)


def test_field_func_pos():
    f = mv.Fieldset(path=os.path.join(PATH, "tuv_pl.grib"))
    g = +f  # should return values unaltered
    assert type(g) == mv.Fieldset
    assert len(g) == 18
    vf = f.values()
    vg = g.values()
    np.testing.assert_allclose(vg, vf)


def test_field_func_abs():
    f = mv.Fieldset(path=os.path.join(PATH, "tuv_pl.grib"))
    g = f.abs()
    assert type(g) == mv.Fieldset
    assert len(g) == 18
    vf = f.values()
    vg = g.values()
    np.testing.assert_allclose(vg, np.abs(vf))


def test_temporary_file():
    # create a temp file, then delete the fieldset - temp should be removed
    f = mv.Fieldset(path=os.path.join(PATH, "tuv_pl.grib"))
    g = -f
    temp_path = g.temporary.path
    assert os.path.isfile(temp_path)
    g = None
    assert not os.path.isfile(temp_path)


def test_permanent_file_not_accidentally_deleted():
    path = os.path.join(PATH, "tuv_pl.grib")
    f = mv.Fieldset(path=path)
    assert os.path.isfile(path)
    f = None
    assert os.path.isfile(path)


def test_single_index_0():
    path = os.path.join(PATH, "tuv_pl.grib")
    f = mv.Fieldset(path=path)
    f0 = f[0]
    assert type(f0) is mv.Fieldset
    assert len(f0) == 1
    assert f0.grib_get_string("shortName") == "t"
    v = f0.values()
    eps = 0.001
    assert len(v) == 2664
    assert np.isclose(v[1088], 304.5642, eps)


def test_single_index_17():
    path = os.path.join(PATH, "tuv_pl.grib")
    f = mv.Fieldset(path=path)
    f17 = f[17]
    assert type(f17) is mv.Fieldset
    assert len(f17) == 1
    assert f17.grib_get_string("shortName") == "v"
    v = f17.values()
    eps = 0.001
    assert len(v) == 2664
    assert np.isclose(v[2663], -11.0797, eps)


def test_single_index_minus_1():
    path = os.path.join(PATH, "tuv_pl.grib")
    f = mv.Fieldset(path=path)
    fm1 = f[-1]
    assert type(fm1) is mv.Fieldset
    assert len(fm1) == 1
    assert fm1.grib_get_string("shortName") == "v"
    v = fm1.values()
    eps = 0.001
    assert len(v) == 2664
    assert np.isclose(v[2663], -11.0797, eps)


def test_single_index_bad():
    path = os.path.join(PATH, "tuv_pl.grib")
    f = mv.Fieldset(path=path)
    with pytest.raises(IndexError):
        fbad = f[27]


def test_slice_0_5():
    path = os.path.join(PATH, "tuv_pl.grib")
    f = mv.Fieldset(path=path)
    f05 = f[0:5]
    assert type(f05) is mv.Fieldset
    assert len(f05) == 5
    assert f05.grib_get_string("shortName") == ["t", "u", "v", "t", "u"]
    v = f05.values()
    assert v.shape == (5, 2664)
    # check the original fieldset
    assert len(f) == 18
    sn = f.grib_get_string("shortName")
    assert sn == ["t", "u", "v"] * 6


def test_array_indexing():
    path = os.path.join(PATH, "tuv_pl.grib")
    f = mv.Fieldset(path=path)
    indexes = np.array([1, 16, 5, 9])
    fv = f[indexes]
    assert type(fv) is mv.Fieldset
    assert len(fv) == 4
    assert fv.grib_get_string("shortName") == ["u", "u", "v", "t"]
    # check with bad indexes
    indexes = np.array([1, 36, 5, 9])
    with pytest.raises(IndexError):
        fvbad = f[indexes]


def test_fieldset_iterator():
    grib = mv.Fieldset(path=os.path.join(PATH, "tuv_pl.grib"))
    sn = grib.grib_get_string("shortName")
    assert len(sn) == 18
    iter_sn = []
    for f in grib:
        iter_sn.append(f.grib_get_string("shortName"))
    assert len(iter_sn) == len(sn)
    assert iter_sn == sn
    iter_sn = [f.grib_get_string("shortName") for f in grib]
    assert iter_sn == sn


def test_fieldset_iterator_multiple():
    grib = mv.Fieldset(path=os.path.join(PATH, "tuv_pl.grib"))
    sn = grib.grib_get_string("shortName")
    assert len(sn) == 18
    for i in [1, 2, 3]:
        iter_sn = []
        for f in grib:
            iter_sn.append(f.grib_get_string("shortName"))
        assert len(iter_sn) == len(sn)
        for i in range(0, 18):
            assert sn[i] == iter_sn[i]


def test_fieldset_iterator_with_zip():
    # this tests something different with the iterator - this does not try to
    # 'go off the edge' of the fieldset, because the length is determined by
    # the list of levels
    grib = mv.Fieldset(path=os.path.join(PATH, "tuv_pl.grib"))
    ref_levs = grib.grib_get_long("level")
    assert len(ref_levs) == 18
    levs1 = []
    levs2 = []
    for k, f in zip(grib.grib_get_long("level"), grib):
        levs1.append(k)
        levs2.append(f.grib_get_long("level"))
    assert levs1 == ref_levs
    assert levs2 == ref_levs


def test_fieldset_iterator_with_zip_multiple():
    # same as test_fieldset_iterator_with_zip() but multiple times
    grib = mv.Fieldset(path=os.path.join(PATH, "tuv_pl.grib"))
    ref_levs = grib.grib_get_long("level")
    assert len(ref_levs) == 18
    for i in [1, 2, 3]:
        levs1 = []
        levs2 = []
        for k, f in zip(grib.grib_get_long("level"), grib):
            levs1.append(k)
            levs2.append(f.grib_get_long("level"))
        print(grib.grib_get_long("level"))
        assert levs1 == ref_levs
        assert levs2 == ref_levs


def test_fieldset_reverse_iterator():
    grib = mv.Fieldset(path=os.path.join(PATH, "tuv_pl.grib"))
    sn = grib.grib_get_string("shortName")
    sn_reversed = list(reversed(sn))
    assert sn_reversed[0] == "v"
    assert sn_reversed[17] == "t"
    gribr = reversed(grib)
    iter_sn = [f.grib_get_string("shortName") for f in gribr]
    assert len(iter_sn) == len(sn_reversed)
    assert iter_sn == sn_reversed
    assert iter_sn == ["v", "u", "t"] * 6


def test_fieldset_append():
    g = mv.Fieldset(path=os.path.join(PATH, "tuv_pl.grib"))
    h = mv.Fieldset(path=os.path.join(PATH, "all_missing_vals.grib"))
    i = g[0:3]
    i.append(h)
    assert i.grib_get_string("shortName") == ["t", "u", "v", "z"]


def test_fieldset_merge():
    g = mv.Fieldset(path=os.path.join(PATH, "tuv_pl.grib"))
    h = mv.Fieldset(path=os.path.join(PATH, "all_missing_vals.grib"))
    i = g[0:3]
    j = i.merge(h)  # does not alter the original fieldset
    assert i.grib_get_string("shortName") == ["t", "u", "v"]
    assert j.grib_get_string("shortName") == ["t", "u", "v", "z"]


def test_field_scalar_func():
    f = mv.Fieldset(path=os.path.join(PATH, "tuv_pl.grib"))[0:3]
    # fieldset op scalar
    g = f + 10
    assert type(g) == mv.Fieldset
    assert len(g) == 3
    np.testing.assert_allclose(g.values(), f.values() + 10)
    q = f - 5
    np.testing.assert_allclose(q.values(), f.values() - 5)
    m = f * 1.5
    np.testing.assert_allclose(m.values(), f.values() * 1.5)
    d = f / 3.0
    np.testing.assert_allclose(d.values(), f.values() / 3.0, 0.0001)
    p = f**2
    np.testing.assert_allclose(p.values(), f.values() ** 2)
    first_val = f.values()[0][0]  # 272
    ge = f >= first_val
    v = ge.values()
    assert v[0][0] == 1  # 272
    assert v[0][2645] == 0  # 240
    assert v[0][148] == 1  # 280
    assert v[1][0] == 0  # -6
    gt = f > first_val
    v = gt.values()
    assert v[0][0] == 0  # 272
    assert v[0][2645] == 0  # 240
    assert v[0][148] == 1  # 280
    assert v[1][0] == 0  # - 6
    lt = f < first_val
    v = lt.values()
    assert v[0][0] == 0  # 272
    assert v[0][2645] == 1  # 240
    assert v[0][148] == 0  # 280
    assert v[1][0] == 1  # - 6
    lt = f <= first_val
    v = lt.values()
    assert v[0][0] == 1  # 272
    assert v[0][2645] == 1  # 240
    assert v[0][148] == 0  # 280
    assert v[1][0] == 1  # - 6
    e = f == first_val
    v = e.values()
    assert v[0][0] == 1  # 272
    assert v[0][2645] == 0  # 240
    assert v[0][148] == 0  # 280
    assert v[1][0] == 0  # - 6
    ne = f != first_val
    v = ne.values()
    assert v[0][0] == 0  # 272
    assert v[0][2645] == 1  # 240
    assert v[0][148] == 1  # 280
    assert v[1][0] == 1  # - 6
    andd = (f > 270) & (f < 290)  # and
    v = andd.values()
    assert v[0][0] == 1  # 272
    assert v[0][2645] == 0  # 240
    assert v[0][148] == 1  # 280
    assert v[1][0] == 0  # - 6
    orr = (f < 270) | (f > 279)  # or
    v = orr.values()
    assert v[0][0] == 0  # 272
    assert v[0][2645] == 1  # 240
    assert v[0][148] == 1  # 280
    assert v[1][0] == 1  # - 6
    nott = ~((f > 270) & (f < 290))  # not
    v = nott.values()
    assert v[0][0] == 0  # 272
    assert v[0][2645] == 1  # 240
    assert v[0][148] == 0  # 280
    assert v[1][0] == 1  # - 6
    # scalar op fieldset
    h = 20 + f
    assert type(h) == mv.Fieldset
    assert len(h) == 3
    np.testing.assert_allclose(h.values(), f.values() + 20)
    r = 25 - f
    np.testing.assert_allclose(r.values(), 25 - f.values())
    mr = 3 * f
    np.testing.assert_allclose(mr.values(), f.values() * 3)
    dr = 200 / f
    np.testing.assert_allclose(dr.values(), 200 / f.values(), 0.0001)
    pr = 2**f
    np.testing.assert_allclose(pr.values(), 2 ** f.values(), 1)


def test_fieldset_fieldset_func():
    f = mv.Fieldset(path=os.path.join(PATH, "tuv_pl.grib"))[0:3]
    g = mv.Fieldset(path=os.path.join(PATH, "tuv_pl.grib"))[5:8]
    h = f + g
    np.testing.assert_allclose(h.values(), f.values() + g.values())
    i = g + f
    np.testing.assert_allclose(i.values(), g.values() + f.values())
    q = f - g
    np.testing.assert_allclose(q.values(), f.values() - g.values())
    r = g - f
    np.testing.assert_allclose(r.values(), g.values() - f.values())
    t = g * f
    np.testing.assert_allclose(t.values(), g.values() * f.values(), 0.0001)
    d = g / f
    np.testing.assert_allclose(d.values(), g.values() / f.values(), 0.0001)
    gt = f > g
    assert gt[0].values()[0] == 1
    assert gt[1].values()[0] == 0
    assert gt[2].values()[0] == 1
    assert gt[2].values()[22] == 0
    gt = f >= g
    assert gt[0].values()[0] == 1
    assert gt[1].values()[0] == 0
    assert gt[2].values()[0] == 1
    assert gt[2].values()[22] == 0
    lt = f < g
    assert lt[0].values()[0] == 0
    assert lt[1].values()[0] == 1
    assert lt[2].values()[0] == 0
    assert lt[2].values()[22] == 1
    lt = f <= g
    assert lt[0].values()[0] == 0
    assert lt[1].values()[0] == 1
    assert lt[2].values()[0] == 0
    assert lt[2].values()[22] == 1


def test_fieldset_multiple_funcs():
    f = mv.Fieldset(path=os.path.join(PATH, "tuv_pl.grib"))
    g = 1 - ((f[0] + f[3]) - 5)
    np.testing.assert_allclose(g.values(), 1 - ((f[0].values() + f[3].values()) - 5))


def test_fieldset_funs_with_read():
    f = mv.read(os.path.join(PATH, "tuv_pl.grib"))
    assert isinstance(f, mv.Fieldset)
    g = f + 18
    assert isinstance(g, mv.Fieldset)
    diff = g - f
    assert isinstance(diff, mv.Fieldset)
    assert np.isclose(diff.minvalue(), 18)
    assert np.isclose(diff.maxvalue(), 18)


def test_field_maths_funcs():
    f = mv.Fieldset(path=os.path.join(PATH, "tuv_pl.grib"))
    f = f[0]
    v = f.values()

    # no arg
    r = f.abs()
    np.testing.assert_allclose(r.values(), np.fabs(v), rtol=1e-05)

    r = f.cos()
    np.testing.assert_allclose(r.values(), np.cos(v), rtol=1e-05)

    f1 = f / 100
    r = f1.exp()
    np.testing.assert_allclose(r.values(), np.exp(f1.values()), rtol=1e-05)

    r = f.log()
    np.testing.assert_allclose(r.values(), np.log(v), rtol=1e-05)

    r = f.log10()
    np.testing.assert_allclose(r.values(), np.log10(v), rtol=1e-05)

    r = f.sgn()
    np.testing.assert_allclose(r.values(), np.sign(v), rtol=1e-05)

    r = f.sin()
    np.testing.assert_allclose(r.values(), np.sin(v), rtol=1e-05)

    r = f.sqrt()
    np.testing.assert_allclose(r.values(), np.sqrt(v), rtol=1e-05)

    r = f.tan()
    np.testing.assert_allclose(r.values(), np.tan(v), rtol=1e-04)

    # inverse functions
    # scale input between [-1, 1]
    f1 = (f - 282) / 80
    v1 = f1.values()
    r = f1.acos()
    np.testing.assert_allclose(r.values(), np.arccos(v1), rtol=1e-05)

    r = f1.asin()
    np.testing.assert_allclose(r.values(), np.arcsin(v1), rtol=1e-05)

    r = f1.atan()
    np.testing.assert_allclose(r.values(), np.arctan(v1), rtol=1e-05)

    # 1 arg
    f1 = f - 274
    v1 = f1.values()

    r = f.atan2(f1)
    np.testing.assert_allclose(r.values(), np.arctan2(v, v1), rtol=1e-05)

    r = f.div(f1)
    np.testing.assert_allclose(r.values(), np.floor_divide(v, v1), rtol=1e-05)

    r = f.mod(f1)
    np.testing.assert_allclose(r.values(), np.mod(v, v1), rtol=1e-04)


def test_str():
    f = mv.Fieldset(path=os.path.join(PATH, "tuv_pl.grib"))
    assert str(f) == "Fieldset (18 fields)"


def test_set_values_single_field():
    f = mv.Fieldset(path=os.path.join(PATH, "tuv_pl.grib"))
    f0 = f[0]
    f0_vals = f0.values()
    vals_plus_10 = f0_vals + 10
    f0_modified = f0.set_values(vals_plus_10)
    f0_mod_vals = f0_modified.values()
    np.testing.assert_allclose(f0_mod_vals, vals_plus_10)
    # write to disk, read and check again
    testpath = "f0_modified.grib"
    f0_modified.write(testpath)
    f0_read = mv.Fieldset(path=testpath)
    np.testing.assert_allclose(f0_read.values(), vals_plus_10)
    os.remove(testpath)


def test_set_values_multiple_fields():
    f = mv.Fieldset(path=os.path.join(PATH, "tuv_pl.grib"))
    f03 = f[0:3]
    f47 = f[4:7]
    f03_modified = f03.set_values(f47.values())
    np.testing.assert_allclose(f03_modified.values(), f47.values())
    # same, but with a list of arrays instead of a 2D array
    list_of_arrays = [f.values() for f in f47]
    f03_modified_2 = f03.set_values(list_of_arrays)
    np.testing.assert_allclose(f03_modified_2.values(), f47.values())
    # wrong number of arrays
    f48 = f[4:8]
    with pytest.raises(ValueError):
        f03_modified_3 = f03.set_values(f48.values())


def test_set_values_with_missing_values():
    f = mv.Fieldset(path=os.path.join(PATH, "t_with_missing.grib"))
    new_vals = f.values() + 40
    g = f.set_values(new_vals)
    v = g.values()
    assert v.shape == (2664,)
    eps = 0.001
    assert np.isclose(v[0], 272.5642 + 40, eps)
    assert np.isnan(v[798])
    assert np.isnan(v[806])
    assert np.isnan(v[1447])
    assert np.isclose(v[2663], 240.5642 + 40, eps)


def test_set_values_with_missing_values_2():
    f = mv.Fieldset(path=os.path.join(PATH, "t_with_missing.grib"))
    g = f[0]
    v = g.values()
    v[1] = np.nan
    h = g.set_values(v)
    hv = h.values()[:10]
    assert np.isclose(hv[0], 272.56417847)
    assert np.isnan(hv[1])
    assert np.isclose(hv[2], 272.56417847)


def test_set_values_resize():
    # NOTE: the current change in behavour - in 'standard Metview' the user
    # has to supply "resize" as an optional argument in order to allow an array
    # of different size to be used; if not supplied, and the given array is not the
    # same size as the original field, an error is thrown; here, we allow resizing
    # without the need for an extra argument - do we want to do this check?
    f = mv.Fieldset(path=os.path.join(PATH, "tuv_pl.grib"))
    f0 = f[0]
    f0_20vals = f0.values()[0:20]
    f0_modified = f0.set_values(f0_20vals)
    f0_mod_vals = f0_modified.values()
    eps = 0.001
    np.testing.assert_allclose(f0_mod_vals, f0_20vals, eps)


def test_vals_destroyed():
    f = mv.Fieldset(path=os.path.join(PATH, "test.grib"))
    assert f.fields[0].vals is None
    g = f.values()
    assert isinstance(g, np.ndarray)
    assert f.fields[0].vals is None
    f = -f
    assert f.fields[0].vals is None
    g = f.values()
    assert isinstance(g, np.ndarray)
    assert f.fields[0].vals is None
    f = f + 1
    assert f.fields[0].vals is None
    g = f.values()
    assert isinstance(g, np.ndarray)
    assert f.fields[0].vals is None


def test_accumulate():
    f = mv.Fieldset(path=os.path.join(PATH, "t1000_LL_7x7.grb"))
    v = mv.accumulate(f)
    assert isinstance(v, float)
    assert np.isclose(v, 393334.244141)

    f = mv.Fieldset(path=os.path.join(PATH, "monthly_avg.grib"))
    v = mv.accumulate(f)
    assert isinstance(v, list)
    v_ref = [
        408058.256226,
        413695.059631,
        430591.282776,
        428943.981812,
        422329.622498,
        418016.024231,
        409755.097961,
        402741.786194,
    ]
    assert len(v) == len(f)
    np.testing.assert_allclose(v, v_ref)


def test_average():
    fs = mv.Fieldset(path=os.path.join(PATH, "test.grib"))

    # const fields
    v = mv.average(fs * 0 + 1)
    assert isinstance(v, float)
    assert np.isclose(v, 1)

    # # single field
    v = mv.average(fs)
    assert isinstance(v, float)
    assert np.isclose(v, 279.06647863)

    # multiple fields
    f = mv.Fieldset(path=os.path.join(PATH, "monthly_avg.grib"))
    v = mv.average(f)
    assert isinstance(v, list)
    v_ref = [
        290.639783636,
        294.654600877,
        306.688947846,
        305.515656561,
        300.804574428,
        297.732210991,
        291.848360371,
        286.85312407,
    ]

    assert len(v) == len(f)
    np.testing.assert_allclose(v, v_ref)


def test_latitudes():
    fs = mv.Fieldset(path=os.path.join(PATH, "t1000_LL_2x2.grb"))

    v = mv.latitudes(fs)
    assert isinstance(v, np.ndarray)
    assert len(v) == 16380
    assert np.isclose(v[0], 90)
    assert np.isclose(v[1], 90)
    assert np.isclose(v[8103], 0)
    assert np.isclose(v[11335], -34)
    assert np.isclose(v[16379], -90)

    f = fs.merge(fs)
    lst = mv.latitudes(f)
    assert len(lst) == 2
    for v in lst:
        assert np.isclose(v[0], 90)
        assert np.isclose(v[1], 90)
        assert np.isclose(v[8103], 0)
        assert np.isclose(v[11335], -34)
        assert np.isclose(v[16379], -90)


def test_longitudes():
    fs = mv.Fieldset(path=os.path.join(PATH, "t1000_LL_2x2.grb"))

    v = mv.longitudes(fs)
    assert isinstance(v, np.ndarray)
    assert len(v) == 16380
    assert np.isclose(v[0], 0)
    assert np.isclose(v[1], 2)
    assert np.isclose(v[8103], 6)
    assert np.isclose(v[11335], 350)
    assert np.isclose(v[16379], 358)

    f = fs.merge(fs)
    lst = mv.longitudes(f)
    assert len(lst) == 2
    for v in lst:
        assert np.isclose(v[0], 0)
        assert np.isclose(v[1], 2)
        assert np.isclose(v[8103], 6)
        assert np.isclose(v[11335], 350)
        assert np.isclose(v[16379], 358)


def test_coslat():
    fs = mv.Fieldset(path=os.path.join(PATH, "t_time_series.grib"))

    # WARN: it is important that the data should be at least 16 bit
    #  to keep accuracy in resulting fields

    f = fs[0]
    r = mv.coslat(f)
    np.testing.assert_allclose(
        r.values(), np.cos(np.deg2rad(f.latitudes())), rtol=1e-06
    )

    f = fs[:2]
    r = mv.coslat(f)
    assert len(r) == 2
    for i in range(len(r)):
        np.testing.assert_allclose(
            r[i].values(), np.cos(np.deg2rad(f[i].latitudes())), rtol=1e-06
        )


def test_mean():
    fs = mv.Fieldset(path=os.path.join(PATH, "test.grib"))

    # single fields
    f = fs
    r = mv.mean(f)
    v_ref = mv.values(fs)
    assert len(r) == 1
    np.testing.assert_allclose(r.values(), v_ref, rtol=1e-05)

    # known mean
    f = fs.merge(2 * fs)
    f = f.merge(3 * fs)
    r = f.mean()
    v_ref = mv.values(fs) * 2
    assert len(r) == 1
    np.testing.assert_allclose(r.values(), v_ref, rtol=1e-05)


def test_maxvalue():
    fs = mv.Fieldset(path=os.path.join(PATH, "test.grib"))

    f = fs
    f = f.merge(3 * fs)
    f = f.merge(2 * fs)
    v = mv.maxvalue(f)
    assert isinstance(v, float)
    assert np.isclose(v, 948.1818237304688)


def test_minvalue():
    fs = mv.Fieldset(path=os.path.join(PATH, "test.grib"))

    f = 3 * fs
    f = f.merge(fs)
    f = f.merge(2 * fs)
    v = mv.minvalue(f)
    assert isinstance(v, float)
    assert np.isclose(v, 206.93560791015625)


def test_sinlat():
    fs = mv.Fieldset(path=os.path.join(PATH, "t_time_series.grib"))

    # WARN: it is important that the data should be at least 16 bit
    #  to keep accuracy in resulting fields

    f = fs[0]
    r = mv.sinlat(f)
    np.testing.assert_allclose(
        r.values(), np.sin(np.deg2rad(f.latitudes())), rtol=1e-06
    )

    f = fs[:2]
    r = mv.sinlat(f)
    assert len(r) == 2
    for i in range(len(r)):
        np.testing.assert_allclose(
            r[i].values(), np.sin(np.deg2rad(f[i].latitudes())), rtol=1e-06
        )


def test_tanlat():
    fs = mv.Fieldset(path=os.path.join(PATH, "t_time_series.grib"))

    # WARN: it is important that the data should be at least 16 bit
    #  to keep accuracy in resulting fields

    # TODO: use pole_limit value from fieldset

    pole_limit = 90.0 - 1e-06

    f = fs[0]
    r = mv.tanlat(f)
    lat = f.latitudes()
    lat[np.fabs(lat) > pole_limit] = np.nan
    np.testing.assert_allclose(
        r.values(), np.tan(np.deg2rad(lat)), rtol=1e-06, atol=1e-06
    )

    f = fs[:2]
    r = mv.tanlat(f)
    assert len(r) == 2
    for i in range(len(r)):
        lat = f[i].latitudes()
        lat[np.fabs(lat) > pole_limit] = np.nan
        np.testing.assert_allclose(
            r[i].values(), np.tan(np.deg2rad(lat)), rtol=1e-06, atol=1e-06
        )


def test_stdev():
    fs = mv.Fieldset(path=os.path.join(PATH, "t1000_LL_7x7.grb"))

    # single field
    r = mv.stdev(fs)
    assert len(r) == 1
    np.testing.assert_allclose(r.values(), fs.values() * 0)

    # known variance
    f = fs.merge(4 * fs)
    f = f.merge(10 * fs)
    r = mv.stdev(f)
    assert len(r) == 1
    np.testing.assert_allclose(r.values(), np.sqrt(np.square(fs.values()) * 42 / 3))

    # real life example
    fs = mv.Fieldset(path=os.path.join(PATH, "monthly_avg.grib"))
    r = mv.stdev(fs)
    assert len(r) == 1

    v_ref = np.ma.std(np.array([x.values() for x in fs]), axis=0)
    np.testing.assert_allclose(r.values(), v_ref, rtol=1e-03)


def test_sum():
    fs = mv.Fieldset(path=os.path.join(PATH, "t1000_LL_7x7.grb"))

    # single fields
    f = fs
    r = mv.sum(f)
    assert len(r) == 1
    np.testing.assert_allclose(r.values(), fs.values())

    # known sum
    f = fs.merge(fs)
    f = f.merge(fs)
    r = f.sum()
    assert len(r) == 1
    np.testing.assert_allclose(r.values(), fs.values() * 3)

    # real life example
    f = mv.Fieldset(path=os.path.join(PATH, "monthly_avg.grib"))
    r = f.sum()
    assert len(r) == 1
    v_ref = r.values() * 0
    for g in f:
        v_ref += g.values()
    np.testing.assert_allclose(r.values(), v_ref)


def test_var():
    fs = mv.Fieldset(path=os.path.join(PATH, "t1000_LL_7x7.grb"))

    # single field
    r = mv.var(fs)
    assert len(r) == 1
    np.testing.assert_allclose(r.values(), fs.values() * 0)

    # known variance
    f = fs.merge(4 * fs)
    f = f.merge(10 * fs)
    r = mv.var(f)
    assert len(r) == 1
    np.testing.assert_allclose(r.values(), np.square(fs.values()) * 42 / 3)

    # real life example
    fs = mv.Fieldset(path=os.path.join(PATH, "monthly_avg.grib"))
    r = mv.var(fs)
    assert len(r) == 1

    v_ref = np.ma.var(np.array([x.values() for x in fs]), axis=0)
    np.testing.assert_allclose(r.values(), v_ref, rtol=1e-03)


def test_date():

    fs = mv.Fieldset(path=os.path.join(PATH, "monthly_avg.grib"))

    # analysis, so valid=base
    bdate_ref = [
        "2016-01-01 00:00:00",
        "2016-02-01 00:00:00",
        "2016-03-01 00:00:00",
        "2016-04-01 00:00:00",
        "2016-05-01 00:00:00",
        "2016-06-01 00:00:00",
        "2016-07-01 00:00:00",
        "2016-08-01 00:00:00",
    ]
    vdate_ref = bdate_ref

    v = mv.base_date(fs)
    assert len(v) == len(fs)
    for i, d in enumerate(v):
        assert d == utils.date_from_str(bdate_ref[i])

    v = mv.valid_date(fs)
    assert len(v) == len(fs)
    for i, d in enumerate(v):
        assert d == utils.date_from_str(vdate_ref[i])


def test_bitmap():
    fs = mv.Fieldset(path=os.path.join(PATH, "t1000_LL_2x2.grb"))

    # -- const field
    f = fs * 0 + 1

    # non missing
    r = mv.bitmap(f, 0)
    np.testing.assert_allclose(r.values(), f.values())

    # all missing
    r = mv.bitmap(f, 1)
    np.testing.assert_allclose(r.values(), f.values() * np.nan)

    # -- non const field
    f = fs

    # bitmap with value
    f_mask = f > 300
    r = mv.bitmap(f_mask, 1)
    v_ref = f_mask.values()
    v_ref[v_ref == 1] = np.nan
    np.testing.assert_allclose(r.values(), v_ref)

    f_mask = f > 300
    r = mv.bitmap(f_mask * 2, 2)
    v_ref = f_mask.values() * 2
    v_ref[v_ref == 2] = np.nan
    np.testing.assert_allclose(r.values(), v_ref)

    # bitmap with field
    f = mv.bitmap(fs > 300, 0)
    r = mv.bitmap(fs, f)
    v_ref = fs.values() * f.values()
    np.testing.assert_allclose(r.values(), v_ref)

    # multiple fields
    f = mv.Fieldset(path=os.path.join(PATH, "monthly_avg.grib"))
    f = f[0:2]

    # with value
    f_mask = f > 300
    r = mv.bitmap(f_mask, 1)
    assert len(r) == len(f)
    for i in range(len(r)):
        v_ref = f_mask[i].values()
        v_ref[v_ref == 1] = np.nan
        np.testing.assert_allclose(r[i].values(), v_ref)

    # with field
    f1 = mv.bitmap(f > 300, 0)
    r = mv.bitmap(f, f1)
    assert len(r) == len(f1)
    for i in range(len(r)):
        v_ref = f[i].values() * f1[i].values()
        np.testing.assert_allclose(r[i].values(), v_ref)

    # with single field
    f1 = mv.bitmap(f[0] > 300, 0)
    r = mv.bitmap(f, f1)
    assert len(r) == len(f)
    for i in range(len(r)):
        v_ref = f[i].values() * f1.values()
        np.testing.assert_allclose(r[i].values(), v_ref)


def test_nobitmap():

    fs = mv.Fieldset(path=os.path.join(PATH, "t_with_missing.grib"))

    # single field
    f = fs
    r = mv.nobitmap(f, 1)
    assert len(r) == 1
    v_ref = f.values()
    v_ref[np.isnan(v_ref)] = 1
    np.testing.assert_allclose(r.values(), v_ref)

    # multiple fields
    f = fs.merge(2 * fs)
    r = mv.nobitmap(f, 1)
    assert len(r) == 2

    for i in range(len(r)):
        v_ref = f[i].values()
        v_ref[np.isnan(v_ref)] = 1
        np.testing.assert_allclose(r[i].values(), v_ref)


def test_grib_index_0():
    # empty fieldset
    fs = mv.Fieldset()
    gi = fs.grib_index()
    assert gi == []


def test_grib_index_1():
    # single field
    grib_path = os.path.join(PATH, "test.grib")
    fs = mv.Fieldset(path=grib_path)
    gi = fs.grib_index()
    assert gi == [(grib_path, 0)]


def test_grib_index_2():
    # multiple fields
    grib_path = os.path.join(PATH, "tuv_pl.grib")
    fs = mv.Fieldset(path=grib_path)
    gi = fs.grib_index()
    assert isinstance(gi, list)
    assert len(gi) == 18
    for f, g in zip(fs, gi):
        assert g == (grib_path, f.grib_get_long("offset"))
    assert gi[5] == (grib_path, 7200)


def test_grib_index_3():
    # merged fields from different files
    gp1 = os.path.join(PATH, "tuv_pl.grib")
    gp2 = os.path.join(PATH, "t_time_series.grib")
    fs1 = mv.Fieldset(path=gp1)
    fs2 = mv.Fieldset(path=gp2)
    fs3 = fs1[4:7]
    fs3.append(fs2[1])
    fs3.append(fs1[2])
    gi = fs3.grib_index()
    assert isinstance(gi, list)
    assert len(gi) == 5
    # assert gi == [(gp1, 5760), (gp1, 7200), (gp1, 8640), (gp2, 5520), (gp1, 2880)]
    assert gi == [(gp1, 5760), (gp1, 7200), (gp1, 8640), (gp2, 5436), (gp1, 2880)]


def test_grib_index_4():
    # test with a derived fieldset
    fs = mv.Fieldset(os.path.join(PATH, "t_time_series.grib"))[0:4]
    fs1 = fs + 1
    gi = fs1.grib_index()
    for i in gi:
        assert is_temp_file(i[0])
    offsets = [i[1] for i in gi]
    assert offsets == [0, 8440, 16880, 25320]


def test_grib_index_5():
    # test with grib written with write() function
    f_orig = mv.Fieldset(path=os.path.join(PATH, "tuv_pl.grib"))
    f = (f_orig[0:4]).merge(f_orig[7])
    p = "written_tuv_pl_5.grib"
    f.write(p)
    gi = f.grib_index()
    assert gi == [(p, 0), (p, 1440), (p, 2880), (p, 4320), (p, 5760)]
    f = 0
    os.remove(p)
